# Copyright 2008 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import chirp_common, util, errors, bitwise
from chirp.memmap import MemoryMapBytes

LOG = logging.getLogger(__name__)

TUNING_STEPS = [
    5.0, 6.25, 8.33,  9.0, 10.0, 12.5, 15, 20, 25, 30, 50, 100, 125, 200
    ]

MODES = ["FM", "NFM", "WFM", "AM", "DV"]
DUPLEX = ["", "-", "+"]
TMODES = ["", "Tone", "TSQL", "TSQL", "DTCS", "DTCS"]
DTCS_POL = ["NN", "NR", "RN", "RR"]

MEM_LEN = 34
DV_MEM_LEN = 60


# Dirty hack until I clean up this IC9x mess
class IC9xMemory(chirp_common.Memory):
    """A dirty hack to stash bank information in a memory"""
    _bank = None
    _bank_index = 0

    def __init__(self):
        chirp_common.Memory.__init__(self)


class IC9xDVMemory(chirp_common.DVMemory):
    """See above dirty hack"""
    _bank = None
    _bank_index = 0

    def __init__(self):
        chirp_common.DVMemory.__init__(self)


def _ic9x_parse_frames(buf):
    frames = []

    while b"\xfe\xfe" in buf:
        try:
            start = buf.index(b"\xfe\xfe")
            end = buf[start:].index(b"\xfd") + start + 1
        except Exception:
            LOG.error("No trailing bit")
            break

        framedata = buf[start:end]
        buf = buf[end:]

        try:
            frame = IC92Frame()
            frame.from_raw(framedata[2:-1])
            frames.append(frame)
        except errors.InvalidDataError as e:
            LOG.error("Broken frame: %s" % e)

        # LOG.debug("Parsed %i frames" % len(frames))

    return frames


def ic9x_recv(pipe):
    data = b""
    while b'\xfd' not in data:
        buf = pipe.read(1)
        if not buf:
            break

        data += buf

    return _ic9x_parse_frames(data)


def ic9x_send(pipe, buf):
    """Send @buf to @pipe, wrapped in a header and trailer.  Attempt to read
    any response frames, which are returned as a list"""

    # Add header and trailer
    realbuf = b"\xfe\xfe" + buf + b"\xfd"

    # LOG.debug("Sending:\n%s" % util.hexprint(realbuf))

    pipe.write(realbuf)
    pipe.flush()
    return ic9x_recv(pipe)


class IC92Frame:
    """IC9x frame base class"""
    def get_vfo(self):
        """Return the vfo number"""
        return ord(self._map[0])

    def set_vfo(self, vfo):
        """Set the vfo number"""
        self._map[0] = vfo

    def from_raw(self, data):
        """Construct the frame from raw data"""
        self._map = MemoryMapBytes(data)

    def from_frame(self, frame):
        """Construct the frame by copying another frame"""
        self._map = MemoryMapBytes(frame.get_raw())

    def __init__(self, subcmd=0, flen=0, cmd=0x1A):
        self._map = MemoryMapBytes(b"\x00" * (4 + flen))
        self._map[0] = b"\x01\x80" + bytes([cmd, subcmd])

    def get_payload(self):
        """Return the entire payload (sans header)"""
        return MemoryMapBytes(self._map[4:])

    def get_raw(self):
        """Return the raw version of the frame"""
        return self._map.get_byte_compatible().get_packed()

    def __str__(self):
        string = "Frame VFO=%i (len = %i)\n" % (self.get_vfo(),
                                                len(self.get_payload()))
        string += util.hexprint(self.get_payload())
        string += "\n"

        return string

    def send(self, pipe, verbose=False):
        """Send the frame to the radio via @pipe"""
        if verbose:
            LOG.debug("Sending:\n%s" % util.hexprint(self.get_raw()))

        response = ic9x_send(pipe, self.get_raw())
        while (response and len(response[0].get_raw()) > 4 and
               response[0].get_raw()[4] != self.get_raw()[4]):
            LOG.warning('Skipping unexpected frame:\n%s',
                        util.hexprint(response[0].get_raw()))
            response = ic9x_recv(pipe)

        if not response:
            raise errors.InvalidDataError("No response from radio")

        return response[0]

    def __setitem__(self, start, value):
        self._map[start+4] = value

    def __getitem__(self, index):
        return self.get_payload()[index]

    def __getslice__(self, start, end):
        return self._map[start+4:end+4]


class IC92GetBankFrame(IC92Frame):
    """A frame for requesting bank information"""
    def __init__(self):
        IC92Frame.__init__(self, 0x09)

    def send(self, pipe, verbose=False):
        rframes = ic9x_send(pipe, self.get_raw())

        if len(rframes) == 0:
            raise errors.InvalidDataError("No response from radio")

        return rframes


class IC92BankFrame(IC92Frame):
    """A frame for bank information"""
    def __init__(self):
        # 1 byte for identifier
        # 8 bytes for name
        IC92Frame.__init__(self, 0x0B, 9)

    def get_name(self):
        """Return the bank name"""
        return self[1:]

    def get_identifier(self):
        """Return the letter for the bank (A-Z)"""
        return self[0]

    def set_name(self, name):
        """Set the bank name"""
        self[1] = name[:8].ljust(8)

    def set_identifier(self, ident):
        """Set the letter for the bank (A-Z)"""
        self[0] = ident[0]


class IC92MemClearFrame(IC92Frame):
    """A frame for clearing (erasing) a memory"""
    def __init__(self, loc):
        # 2 bytes for location
        # 1 byte for 0xFF
        IC92Frame.__init__(self, 0x00, 4)

        self[0] = struct.pack(">BHB", 1, int("%i" % loc, 16), 0xFF)


class IC92MemGetFrame(IC92Frame):
    """A frame for requesting a memory"""
    def __init__(self, loc, iscall=False):
        # 2 bytes for location
        IC92Frame.__init__(self, 0x00, 3)

        if iscall:
            call = 2
        else:
            call = 1

        self[0] = struct.pack(">BH", call, int("%i" % loc, 16))


class IC92GetCallsignFrame(IC92Frame):
    """A frame for getting callsign information"""
    def __init__(self, calltype, number):
        IC92Frame.__init__(self, calltype, 1, 0x1D)

        self[0] = chr(number)


class IC92CallsignFrame(IC92Frame):
    """A frame to communicate callsign information"""
    command = 0  # Invalid
    width = 8

    def __init__(self, number=0, callsign=""):
        # 1 byte for index
        # $width bytes for callsign
        IC92Frame.__init__(self, self.command, self.width+1, 0x1D)

        self[0] = chr(number) + callsign[:self.width].ljust(self.width)

    def get_callsign(self):
        """Return the actual callsign"""
        return self[1:self.width+1].rstrip()


class IC92YourCallsignFrame(IC92CallsignFrame):
    """URCALL frame"""
    command = 6  # Your


class IC92RepeaterCallsignFrame(IC92CallsignFrame):
    """RPTCALL frame"""
    command = 7  # Repeater


class IC92MyCallsignFrame(IC92CallsignFrame):
    """MYCALL frame"""
    command = 8  # My
    width = 12  # 4 bytes for /STID


MEMORY_FRAME_FORMAT = """
struct {
  u8 vfo;
  bbcd number[2];
  lbcd freq[5];
  lbcd offset[4];
  u8 unknown8;
  bbcd rtone[2];
  bbcd ctone[2];
  bbcd dtcs[2];
  u8 unknown9[2];
  u8 unknown2:1,
     mode:3,
     tuning_step:4;
  u8 unknown1:3,
     tmode: 3,
     duplex: 2;
  u8 unknown5:4,
     dtcs_polarity:2,
     pskip:1,
     skip:1;
  char bank;
  bbcd bank_index[1];
  char name[8];
  u8 unknown10;
  u8 digital_code;
  char rpt2call[8];
  char rpt1call[8];
  char urcall[8];
} mem[1];
"""


class IC92MemoryFrame(IC92Frame):
    """A frame for communicating memory information"""
    def __init__(self):
        IC92Frame.__init__(self, 0, DV_MEM_LEN)

        # For good measure, here is a whole, valid memory block
        # at 146.010 FM.  Since the 9x will complain if any bits
        # are invalid, it's easiest to start with a known-good one
        # since we don't set everything.
        self[0] = \
            b"\x01\x00\x03\x00\x00\x01\x46\x01" + \
            b"\x00\x00\x60\x00\x00\x08\x85\x08" + \
            b"\x85\x00\x23\x22\x80\x06\x00\x00" + \
            b"\x00\x00\x20\x20\x20\x20\x20\x20" + \
            b"\x20\x20\x00\x00\x20\x20\x20\x20" + \
            b"\x20\x20\x20\x20\x4b\x44\x37\x52" + \
            b"\x45\x58\x20\x43\x43\x51\x43\x51" + \
            b"\x43\x51\x20\x20"

    def set_vfo(self, vfo):
        IC92Frame.set_vfo(self, vfo)
        if vfo == 1:
            self._map.truncate(MEM_LEN + 4)

    def set_iscall(self, iscall):
        """This frame refers to a call channel if @iscall is True"""
        if iscall:
            self[0] = 2
        else:
            self[0] = 1

    def get_iscall(self):
        """Return True if this frame refers to a call channel"""
        return ord(self[0]) == 2

    def set_memory(self, mem):
        """Take Memory object @mem and configure the frame accordingly"""
        if mem.number < 0:
            self.set_iscall(True)
            mem.number = abs(mem.number) - 1
            LOG.debug("Memory is %i (call %s)" %
                      (mem.number, self.get_iscall()))

        _mem = bitwise.parse(MEMORY_FRAME_FORMAT, self).mem

        _mem.number = mem.number

        _mem.freq = mem.freq
        _mem.offset = mem.offset
        _mem.rtone = int(mem.rtone * 10)
        _mem.ctone = int(mem.ctone * 10)
        _mem.dtcs = int(mem.dtcs)
        _mem.mode = MODES.index(mem.mode)
        _mem.tuning_step = TUNING_STEPS.index(mem.tuning_step)
        _mem.duplex = DUPLEX.index(mem.duplex)
        _mem.tmode = TMODES.index(mem.tmode)
        _mem.dtcs_polarity = DTCS_POL.index(mem.dtcs_polarity)

        if mem._bank is not None:
            _mem.bank = chr(ord("A") + mem._bank)
            _mem.bank_index = mem._bank_index

        _mem.skip = mem.skip == "S"
        _mem.pskip = mem.skip == "P"

        _mem.name = mem.name.ljust(8)[:8]

        if mem.mode == "DV":
            _mem.urcall = mem.dv_urcall.upper().ljust(8)[:8]
            _mem.rpt1call = mem.dv_rpt1call.upper().ljust(8)[:8]
            _mem.rpt2call = mem.dv_rpt2call.upper().ljust(8)[:8]
            _mem.digital_code = mem.dv_code

    def get_memory(self):
        """Return a Memory object based on the contents of the frame"""
        _mem = bitwise.parse(MEMORY_FRAME_FORMAT, self.get_payload()).mem

        if MODES[_mem.mode] == "DV":
            mem = IC9xDVMemory()
        else:
            mem = IC9xMemory()

        mem.number = int(_mem.number)
        if self.get_iscall():
            mem.number = -1 - mem.number

        mem.freq = int(_mem.freq)
        mem.offset = int(_mem.offset)
        mem.rtone = int(_mem.rtone) / 10.0
        mem.ctone = int(_mem.ctone) / 10.0
        mem.dtcs = int(_mem.dtcs)
        mem.mode = MODES[int(_mem.mode)]
        mem.tuning_step = TUNING_STEPS[int(_mem.tuning_step)]
        mem.duplex = DUPLEX[int(_mem.duplex)]
        mem.tmode = TMODES[int(_mem.tmode)]
        mem.dtcs_polarity = DTCS_POL[int(_mem.dtcs_polarity)]

        if int(_mem.bank) != 0:
            mem._bank = ord(str(_mem.bank)) - ord("A")
            mem._bank_index = int(_mem.bank_index)

        if _mem.skip:
            mem.skip = "S"
        elif _mem.pskip:
            mem.skip = "P"
        else:
            mem.skip = ""

        mem.name = str(_mem.name).rstrip()

        if mem.mode == "DV":
            mem.dv_urcall = str(_mem.urcall).rstrip()
            mem.dv_rpt1call = str(_mem.rpt1call).rstrip()
            mem.dv_rpt2call = str(_mem.rpt2call).rstrip()
            mem.dv_code = int(_mem.digital_code)

        return mem


def get_memory_frame(pipe, vfo, number):
    """Get the memory frame for @vfo and @number via @pipe"""
    if number < 0:
        number = abs(number + 1)
        call = True
    else:
        call = False

    frame = IC92MemGetFrame(number, call)
    frame.set_vfo(vfo)

    return frame.send(pipe)


def get_memory(pipe, vfo, number):
    """Get a memory object for @vfo and @number via @pipe"""
    rframe = get_memory_frame(pipe, vfo, number)

    if len(rframe.get_payload()) < 1:
        raise errors.InvalidMemoryLocation("No response from radio")

    if rframe.get_payload()[3] == b'\xff':
        raise errors.InvalidMemoryLocation("Radio says location is empty")

    mf = IC92MemoryFrame()
    mf.from_frame(rframe)
    if mf.get_memory().number != number:
        raise errors.RadioError('get_memory() wanted %i got %i' % (
            number, mf.get_memory().number))

    return mf.get_memory()


def set_memory(pipe, vfo, memory):
    """Set memory @memory on @vfo via @pipe"""
    frame = IC92MemoryFrame()
    frame.set_memory(memory)
    frame.set_vfo(vfo)

    # LOG.debug("Sending (%i):" % (len(frame.get_raw())))
    # LOG.debug(util.hexprint(frame.get_raw()))

    rframe = frame.send(pipe)

    if rframe.get_raw()[2] != 0xfb:
        raise errors.InvalidDataError("Radio reported error:\n%s" %
                                      util.hexprint(rframe.get_payload()))


def erase_memory(pipe, vfo, number):
    """Erase memory @number on @vfo via @pipe"""
    frame = IC92MemClearFrame(number)
    frame.set_vfo(vfo)

    rframe = frame.send(pipe)
    if rframe.get_raw()[2] != 0xfb:
        raise errors.InvalidDataError("Radio reported error")


def get_banks(pipe, vfo):
    """Get banks for @vfo via @pipe"""
    frame = IC92GetBankFrame()
    frame.set_vfo(vfo)

    rframes = frame.send(pipe)

    if vfo == 1:
        base = 180
    else:
        base = 237

    banks = []

    for i in range(base, base+26):
        bframe = IC92BankFrame()
        bframe.from_frame(rframes[i])

        banks.append(bframe.get_name().rstrip())

    return banks


def set_banks(pipe, vfo, banks):
    """Set banks for @vfo via @pipe"""
    for i in range(0, 26):
        bframe = IC92BankFrame()
        bframe.set_vfo(vfo)
        bframe.set_identifier(chr(i + ord("A")))
        bframe.set_name(banks[i])

        rframe = bframe.send(pipe)
        if rframe.get_payload() != "\xfb":
            raise errors.InvalidDataError("Radio reported error")


def get_call(pipe, cstype, number):
    """Get @cstype callsign @number via @pipe"""
    cframe = IC92GetCallsignFrame(cstype.command, number)
    cframe.set_vfo(2)
    rframe = cframe.send(pipe)

    cframe = IC92CallsignFrame()
    cframe.from_frame(rframe)

    return cframe.get_callsign()


def set_call(pipe, cstype, number, call):
    """Set @cstype @call at position @number via @pipe"""
    cframe = cstype(number, call)
    cframe.set_vfo(2)
    rframe = cframe.send(pipe)

    if rframe.get_payload() != "\xfb":
        raise errors.RadioError("Radio reported error")



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
# Copyright 2012 Tom Hayward <tom@tomh.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import glob
import os
import logging
import sys

from chirp import chirp_common, errors

LOG = logging.getLogger(__name__)


def radio_class_id(cls):
    """Return a unique identification string for @cls"""
    ident = "%s_%s" % (cls.VENDOR, cls.MODEL)
    if cls.VARIANT:
        ident += "_%s" % cls.VARIANT
    ident = ident.replace("/", "_")
    ident = ident.replace(" ", "_")
    ident = ident.replace("(", "")
    ident = ident.replace(")", "")
    return ident


ALLOW_DUPS = False


def enable_reregistrations():
    """Set the global flag ALLOW_DUPS=True, which will enable a driver
    to re-register for a slot in the directory without triggering an
    exception"""
    global ALLOW_DUPS
    if not ALLOW_DUPS:
        LOG.info("driver re-registration enabled")
    ALLOW_DUPS = True


def register(cls):
    """Register radio @cls with the directory"""
    global DRV_TO_RADIO
    ident = radio_class_id(cls)
    if ident in list(DRV_TO_RADIO.keys()):
        if ALLOW_DUPS:
            LOG.warn("Replacing existing driver id `%s'" % ident)
        else:
            raise Exception("Duplicate radio driver id `%s'" % ident)
    elif ALLOW_DUPS:
        LOG.info("Registered %s = %s" % (ident, cls.__name__))
    DRV_TO_RADIO[ident] = cls
    RADIO_TO_DRV[cls] = ident

    if not hasattr(cls, '_DETECTED_MODEL'):
        cls._DETECTED_MODEL = False

    return cls


def detected_by(manager_class):
    """Mark a class as detected by another class

    This means it may not be offered directly as a choice and instead is added
    to @manager_class's DETECTED_MODELS list and should be returned from
    detect_from_serial() when appropriate.
    """
    assert issubclass(manager_class, chirp_common.CloneModeRadio)

    def wrapper(cls):
        assert issubclass(cls, chirp_common.CloneModeRadio)

        cls._DETECTED_MODEL = True
        if manager_class.DETECTED_MODELS is None:
            manager_class.DETECTED_MODELS = []
        manager_class.DETECTED_MODELS.append(cls)
        return cls

    return wrapper


DRV_TO_RADIO = {}
RADIO_TO_DRV = {}
AUX_FORMATS = set()


def register_format(name, pattern, readonly=False):
    """Register a named format and file pattern.

    The name and pattern must not exist in the directory already
    (except together). The name should be something like "CSV" or
    "Icom ICF" and the pattern should be a glob like "*.icf".

    Returns a unique name to go in Radio.FORMATS so the UI knows what
    additional formats a driver can read (and write unless readonly is
    set).
    """
    if (name, pattern) not in [(n, p) for n, p, r in AUX_FORMATS]:
        if name in [x[0] for x in AUX_FORMATS]:
            raise Exception('Duplicate format name %r' % name)
    AUX_FORMATS.add((name, pattern, readonly))
    return name


def get_radio(driver):
    """Get radio driver class by identification string"""
    if driver in DRV_TO_RADIO:
        return DRV_TO_RADIO[driver]
    else:
        raise Exception("Unknown radio type `%s'" % driver)


def get_driver(rclass):
    """Get the identification string for a given class"""
    if rclass in RADIO_TO_DRV:
        return RADIO_TO_DRV[rclass]
    elif rclass.__bases__[0] in RADIO_TO_DRV:
        return RADIO_TO_DRV[rclass.__bases__[0]]
    else:
        raise Exception("Unknown radio type `%s'" % rclass)


# This is a mapping table of radio models that have changed in the past.
# ideally we would never do this, but in the case where a radio was added
# as the wrong model name, or a model has to be split, we need to be able
# to open older files and do something intelligent with them.
MODEL_COMPAT = {
    ('Retevis', 'RT-5R'): ('Retevis', 'RT5R'),
    ('Retevis', 'RT-5RV'): ('Retevis', 'RT5RV'),
    ('Signus', 'XTR-5'): ('Cignus', 'XTR-5'),
}


def get_radio_by_image(image_file):
    """Attempt to get the radio class that owns @image_file"""
    if os.path.exists(image_file):
        with open(image_file, "rb") as f:
            filedata = f.read()
    else:
        filedata = b""

    data, metadata = chirp_common.CloneModeRadio._strip_metadata(filedata)

    for rclass in list(DRV_TO_RADIO.values()):
        if not issubclass(rclass, chirp_common.FileBackedRadio):
            continue

        if not metadata:
            # If no metadata, we do the old thing
            try:
                if rclass.match_model(filedata, image_file):
                    return rclass(image_file)
            except Exception as e:
                LOG.error('Radio class %s failed during detection: %s' % (
                    rclass.__name__, e))
                pass

        meta_vendor = metadata.get('vendor')
        meta_model = metadata.get('model')
        meta_variant = metadata.get('variant')

        meta_vendor, meta_model = MODEL_COMPAT.get((meta_vendor, meta_model),
                                                   (meta_vendor, meta_model))

        # If metadata, then it has to match one of the aliases or the parent
        for alias in rclass.ALIASES + [rclass]:
            if (alias.VENDOR == meta_vendor and alias.MODEL == meta_model and
                    (meta_variant is None or alias.VARIANT == meta_variant)):

                class DynamicRadioAlias(rclass):
                    _orig_rclass = rclass
                    VENDOR = meta_vendor
                    MODEL = meta_model
                    VARIANT = metadata.get('variant')

                return DynamicRadioAlias(image_file)

    if metadata:
        ex = errors.ImageMetadataInvalidModel("Unsupported model %s %s" % (
            metadata.get("vendor"), metadata.get("model")))
        ex.metadata = metadata
        raise ex
    else:
        raise errors.ImageDetectFailed("Unknown file format")


def import_drivers(limit=None):
    frozen = getattr(sys, 'frozen', False)
    if sys.platform == 'win32' and frozen:
        # We are in a frozen win32 build, so we can not glob
        # the driver files, but we do not need to anyway
        import chirp.drivers
        for module in chirp.drivers.__all__:
            try:
                __import__('chirp.drivers.%s' % module)
            except Exception as e:
                print('Failed to import %s: %s' % (module, e))
        return

    # Safe import of everything in chirp/drivers. We need to import them
    # to get them to register, but should not abort if one import fails
    chirp_module_base = os.path.dirname(os.path.abspath(__file__))
    driver_files = glob.glob(os.path.join(chirp_module_base,
                                          'drivers',
                                          '*.py'))
    for driver_file in driver_files:
        module, ext = os.path.splitext(driver_file)
        driver_module = os.path.basename(module)
        if limit and driver_module not in limit:
            continue
        __import__('chirp.drivers.%s' % driver_module)



# Copyright 2023 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import (
    bandplan_iaru_r1,
    bandplan_na,
    bitwise,
    chirp_common,
    directory,
    errors,
    memmap,
    util,
)
from chirp.settings import (
    RadioSetting,
    RadioSettingGroup,
    RadioSettings,
    RadioSettingValueBoolean,
    RadioSettingValueInteger,
    RadioSettingValueList,
)

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0000;
struct {
  lbcd rxfreq[4];      // RX Frequency           0-3
  lbcd txfreq[4];      // TX Frequency           4-7
  u8 rxtone;           // PL/DPL Decode          8
  u8 unused_9;         //                        9
  u8 txtone;           // PL/DPL Encode          A
  u8 unused_b;         //                        B
  u8 compander:1,      // Compander              C
     unused_1:1,       //
     highpower:1,      // Power Level
     wide:1,           // Bandwidth
     unused_2:4;       //
  u8 reserved[3];      // Reserved               D-F
} memory[%d];

#seekto 0x002D;
struct {
  u8 unknown_2d_7:1,   //                        002D
     unknown_2d_6:1,   //
     unknown_2d_5:1,   //
     savem:1,          // Battery Save Set
     save:1,           // Battery Save
     beep:1,           // Beep
     voice:1,          // Voice Prompts
     unused_2d_0:1;    //
  u8 squelch;          // Squelch                002E
  u8 tot;              // Time-out Timer         002F
  u8 channel_4[13];    //                        0030-003C
  u8 unused_3d;        //                        003D
  u8 voxl;             // Vox Level              003E
  u8 voxd;             // Vox Delay              003F
  u8 channel_5[13];    //                        0040-004C
  u8 unknown_3[3];     //                        004D-004F
  u8 channel_6[13];    //                        0050-005C
  u8 unknown_4[3];     //                        005D-005F
  u8 channel_7[13];    //                        0060-006C
  u8 volume;           // Volume                 006D
  u8 tone;             // Tone                   006E
  u8 chnumber;         // Channel                006F
  u8 channel_8[13];    //                        0070-007C
  u8 unknown_7d;       //                        007D
  u8 scan;             // Scan                   007E
  u8 backlight;        // Back Light             007F
  u8 channel_9[13];    //                        0080-008C
  u8 wxchnumber;       // Weather Channel        008D
  u8 wxwarn;           // Weather Warn           008E
  u8 roger;            // End Tone               008F
  u8 channel_10[13];   //                        0090-009C
  u8 unknown_9d;       //                        009D
  u8 keylock;          // Key Lock               009E
  u8 unknown_9f;       //                        009F
  u8 channel_11[13];   //                        00A0-00AC
  u8 unknown_ad;       //                        00AD
  u8 unknown_ae;       //                        00AE
  u8 unknown_af;       //                        00AF
  u8 channel_12[13];   //                        00B0-00BC
  u8 autokeylock;      // Key Lock               00BD
  u8 pfkey_lt;         // Key Set < Long         00BE
  u8 pfkey_gt;         // Key Set > Long         00BB
} settings;

#seekto 0x00AD;
struct {
u8 chnl_8:1,      // Busy Channel Lockout  8     00AD
   chnl_7:1,      // Busy Channel Lockout  7
   chnl_6:1,      // Busy Channel Lockout  6
   chnl_5:1,      // Busy Channel Lockout  5
   chnl_4:1,      // Busy Channel Lockout  4
   chnl_3:1,      // Busy Channel Lockout  3
   chnl_2:1,      // Busy Channel Lockout  2
   chnl_1:1;      // Busy Channel Lockout  1
u8 chnl_16:1,     // Busy Channel Lockout 16     00AE
   chnl_15:1,     // Busy Channel Lockout 15
   chnl_14:1,     // Busy Channel Lockout 14
   chnl_13:1,     // Busy Channel Lockout 13
   chnl_12:1,     // Busy Channel Lockout 12
   chnl_11:1,     // Busy Channel Lockout 11
   chnl_10:1,     // Busy Channel Lockout 10
   chnl_9:1;      // Busy Channel Lockout 09
u8 unused_af_7:1, //                             00AF
   unused_af_6:1, //
   chnl_22:1,     // Busy Channel Lockout 22
   chnl_21:1,     // Busy Channel Lockout 21
   chnl_20:1,     // Busy Channel Lockout 20
   chnl_19:1,     // Busy Channel Lockout 19
   chnl_18:1,     // Busy Channel Lockout 18
   chnl_17:1;     // Busy Channel Lockout 17
} bclo;
"""

CMD_ACK = b"\x06"

BACKLIGHT_LIST = ["Off", "ON", "5S", "10S", "15S", "30S"]
PFKEY_US_LIST = ["None", "Weather", "Warn", "Call", "Monitor"]
PFKEY_EU_LIST = ["None", "Warn", "Call", "Monitor"]
SAVE_LIST = ["Normal", "Super"]
TIMEOUTTIMER_LIST = ["%s seconds" % x for x in range(15, 195, 15)]
VOICE_LIST = ["Off", "English"]
VOXD_LIST = ["0.5", "1.0", "1.5", "2.0", "2.5", "3.0"]
VOXL_LIST = ["OFF"] + ["%s" % x for x in range(1, 10)]

PMR_TONES = tuple(
    set(chirp_common.TONES) - set([69.3, 159.8, 165.5, 171.3, 177.3,
                                   183.5, 189.9, 196.6, 199.5, 206.5,
                                   229.1, 254.1]))

PMR_DTCS_CODES = tuple(
    set(chirp_common.DTCS_CODES) - set([36,  53, 122, 145, 212,
                                        225, 246, 252, 255, 266,
                                        274, 325, 332, 356, 446,
                                        452, 454, 455, 462, 523,
                                        526]))


def _enter_programming_mode(radio):
    serial = radio.pipe

    _magic = radio._magic

    try:
        serial.write(_magic)
        for i in range(1, 5):
            ack = serial.read(1)
            if ack == CMD_ACK:
                break
    except:
        raise errors.RadioError("Error communicating with radio")

    if not ack:
        raise errors.RadioError("No response from radio")
    elif ack != CMD_ACK:
        raise errors.RadioError("Radio refused to enter programming mode")

    try:
        serial.write(b"\x02")
        ident = serial.read(8)
    except:
        raise errors.RadioError("Error communicating with radio")

    # check if ident is OK
    for fp in radio._fingerprint:
        if fp in ident:
            break
    else:
        LOG.debug("Incorrect model ID, got this:\n\n" + util.hexprint(ident))
        raise errors.RadioError("Radio identification failed.")

    try:
        serial.write(CMD_ACK)
        ack = serial.read(1)
    except:
        raise errors.RadioError("Error communicating with radio")

    if ack != CMD_ACK:
        raise errors.RadioError("Radio refused to enter programming mode")


def _exit_programming_mode(radio):
    serial = radio.pipe
    try:
        serial.write(b"E")
    except:
        raise errors.RadioError("Radio refused to exit programming mode")


def _read_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'R', block_addr, block_size)
    expectedresponse = b"W" + cmd[1:]
    LOG.debug("Reading block %04x..." % (block_addr))

    try:
        serial.write(cmd)
        response = serial.read(4 + block_size)
        if response[:4] != expectedresponse:
            raise Exception("Error reading block %04x." % (block_addr))

        block_data = response[4:]

        if block_addr != 0 or radio._ack_1st_block:
            serial.write(CMD_ACK)
            ack = serial.read(1)
    except:
        raise errors.RadioError("Failed to read block at %04x" % block_addr)

    if block_addr != 0 or radio._ack_1st_block:
        if ack != CMD_ACK:
            raise Exception("No ACK reading block %04x." % (block_addr))

    return block_data


def _write_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'W', block_addr, block_size)
    data = radio.get_mmap()[block_addr:block_addr + block_size]

    LOG.debug("Writing Data:")
    LOG.debug(util.hexprint(cmd + data))

    try:
        serial.write(cmd + data)
        if serial.read(1) != CMD_ACK:
            raise Exception("No ACK")
    except:
        raise errors.RadioError("Failed to send block "
                                "to radio at %04x" % block_addr)


def do_download(radio):
    LOG.debug("download")
    _enter_programming_mode(radio)

    data = b""

    status = chirp_common.Status()
    status.msg = "Cloning from radio"

    status.cur = 0
    status.max = radio._memsize

    for addr in range(0, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr + radio.BLOCK_SIZE
        radio.status_fn(status)

        block = _read_block(radio, addr, radio.BLOCK_SIZE)
        data += block

        LOG.debug("Address: %04x" % addr)
        LOG.debug(util.hexprint(block))

    _exit_programming_mode(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"

    _enter_programming_mode(radio)

    status.cur = 0
    status.max = radio._memsize

    for start_addr, end_addr in radio._ranges:
        for addr in range(start_addr, end_addr, radio.BLOCK_SIZE_UP):
            status.cur = addr + radio.BLOCK_SIZE_UP
            radio.status_fn(status)
            _write_block(radio, addr, radio.BLOCK_SIZE_UP)

    _exit_programming_mode(radio)


@directory.register
class RB28Radio(chirp_common.CloneModeRadio):
    """RETEVIS RB28"""
    VENDOR = "Retevis"
    MODEL = "RB28"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False
    BLOCK_SIZE = 0x20
    BLOCK_SIZE_UP = 0x10

    TONES = PMR_TONES
    DTCS_CODES = PMR_DTCS_CODES
    PFKEY_LIST = PFKEY_US_LIST
    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.50),
                    chirp_common.PowerLevel("Low", watts=0.50)]
    VALID_BANDS = [(400000000, 480000000)]

    _magic = b"PHOGR\x0B\x28"
    _fingerprint = [b"P32073" + b"\x02\xFF",
                    b"P32073" + b"\x00\xFF",
                    ]
    _upper = 22
    _mem_params = (_upper,  # number of channels
                   )
    _ack_1st_block = False
    _reserved = True
    _frs = True

    _ranges = [
               (0x0000, 0x0160),
              ]
    _memsize = 0x0160

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = False
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = False
        rf.valid_skips = []
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_duplexes = ["", "off"]
        rf.valid_modes = ["FM", "NFM"]  # 25 kHz, 12.5 kHz.
        rf.valid_tones = self.TONES
        rf.valid_dtcs_codes = self.DTCS_CODES
        rf.memory_bounds = (1, self._upper)
        rf.valid_tuning_steps = [2.5, 5., 6.25, 10., 12.5, 25.]
        rf.valid_bands = self.VALID_BANDS

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT % self._mem_params,
                                     self._mmap)

    def sync_in(self):
        """Download from radio"""
        try:
            data = do_download(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = data
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            do_upload(self)
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def get_memory(self, number):

        mem = chirp_common.Memory()

        mem.number = number

        _mem = self._memobj.memory[number - 1]

        if self._reserved:
            _rsvd = _mem.reserved.get_raw(asbytes=False)

        mem.freq = int(_mem.rxfreq) * 10

        # We'll consider any blank (i.e. 0 MHz frequency) to be empty
        if mem.freq == 0:
            mem.empty = True
            return mem

        if _mem.rxfreq.get_raw(asbytes=False) == "\xFF\xFF\xFF\xFF":
            mem.freq = 0
            mem.empty = True
            return mem

        if int(_mem.rxfreq) == int(_mem.txfreq):
            mem.duplex = ""
            mem.offset = 0
        else:
            mem.duplex = int(_mem.rxfreq) > int(_mem.txfreq) and "-" or "+"
            mem.offset = abs(int(_mem.rxfreq) - int(_mem.txfreq)) * 10

        mem.mode = _mem.wide and "FM" or "NFM"

        if _mem.txtone in [0, 0xFF]:
            txmode = ""
        elif _mem.txtone < 0x27:
            mem.rtone = PMR_TONES[_mem.txtone - 1]
            txmode = "Tone"
        elif _mem.txtone >= 0x27:
            tcode = PMR_DTCS_CODES[_mem.txtone - 0x27]
            mem.dtcs = tcode
            txmode = "DTCS"
        else:
            LOG.warn("Bug: tx_mode is %02x" % _mem.txtone)

        if _mem.rxtone in [0, 0xFF]:
            rxmode = ""
        elif _mem.rxtone < 0x27:
            mem.ctone = PMR_TONES[_mem.rxtone - 1]
            rxmode = "Tone"
        elif _mem.rxtone >= 0x27:
            rcode = PMR_DTCS_CODES[_mem.rxtone - 0x27]
            mem.rx_dtcs = rcode
            rxmode = "DTCS"
        else:
            LOG.warn("Bug: rx_mode is %02x" % _mem.rxtone)

        if txmode == "Tone" and not rxmode:
            mem.tmode = "Tone"
        elif txmode == rxmode and txmode == "Tone" and mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif txmode == rxmode and txmode == "DTCS" and mem.dtcs == mem.rx_dtcs:
            mem.tmode = "DTCS"
        elif rxmode or txmode:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txmode, rxmode)

        mem.power = self.POWER_LEVELS[1 - _mem.highpower]

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSettingValueBoolean(_mem.compander)
        rset = RadioSetting("compander", "Compander", rs)
        mem.extra.append(rset)

        _obj = self._memobj.bclo
        key = "chnl_%i" % (mem.number)
        rs = RadioSettingValueBoolean(getattr(_obj, "chnl_%i" % (mem.number)))
        rset = RadioSetting(key, "Busy Channel Lockout", rs)
        mem.extra.append(rset)

        immutable = []

        if self._frs:
            if mem.number >= 1 and mem.number <= 22:
                # 22 FRS fixed channels
                FRS_FREQ = bandplan_na.ALL_GMRS_FREQS[mem.number - 1]
                mem.freq = FRS_FREQ
                mem.duplex == ''
                mem.offset = 0
                mem.mode = "NFM"
                immutable = ["empty", "freq", "duplex", "offset", "mode"]
                if mem.number >= 8 and mem.number <= 14:
                    mem.power = self.POWER_LEVELS[1]
                    immutable += ["power"]
        elif self._pmr:
            if mem.number >= 1 and mem.number <= 16:
                # 16 PMR fixed channels
                PMR_FREQ = bandplan_iaru_r1.PMR446_FREQS[mem.number - 1]
                mem.freq = PMR_FREQ
                mem.duplex = ''
                mem.offset = 0
                mem.mode = "NFM"
                mem.power = self.POWER_LEVELS[1]
                immutable = ["empty", "freq", "duplex", "offset", "mode",
                             "power"]

        mem.immutable = immutable

        return mem

    def _set_tone(self, _mem, which, value, mode):
        if mode == "Tone":
            val = PMR_TONES.index(value) + 1
        elif mode == "DTCS":
            val = PMR_DTCS_CODES.index(value) + 0x27
        elif mode == "":
            val = 0
        else:
            raise errors.RadioError("Internal error: tmode %s" % mode)

        setattr(_mem, which, val)

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]

        if self._reserved:
            _rsvd = _mem.reserved.get_raw(asbytes=False)

        if mem.empty:
            if self._reserved:
                _mem.set_raw("\xFF" * 13 + _rsvd)
            else:
                _mem.set_raw("\xFF" * (_mem.size() // 8))

            return

        if self._reserved:
            _mem.set_raw("\x00" * 13 + _rsvd)
        else:
            _mem.set_raw("\x00" * 13 + "\xFF\xFF\xFF")

        _mem.rxfreq = mem.freq / 10

        if mem.duplex == "off":
            for i in range(0, 4):
                _mem.txfreq[i].set_raw("\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

        _mem.wide = mem.mode == "FM"

        rxtone = txtone = 0
        rxmode = txmode = ""

        if mem.tmode == "DTCS":
            rxmode = txmode = "DTCS"
            rxtone = txtone = mem.dtcs
        elif mem.tmode and mem.tmode != "Cross":
            rxtone = txtone = mem.tmode == "Tone" and mem.rtone or mem.ctone
            txmode = "Tone"
            rxmode = mem.tmode == "TSQL" and "Tone" or ""
        elif mem.tmode == "Cross":
            txmode, rxmode = mem.cross_mode.split("->", 1)

            if txmode == "DTCS":
                txtone = mem.dtcs
            elif txmode == "Tone":
                txtone = mem.rtone

            if rxmode == "DTCS":
                rxtone = mem.rx_dtcs
            elif rxmode == "Tone":
                rxtone = mem.ctone

        self._set_tone(_mem, "txtone", txtone, txmode)
        self._set_tone(_mem, "rxtone", rxtone, rxmode)

        _mem.highpower = mem.power == self.POWER_LEVELS[0]

        for setting in mem.extra:
            if setting.get_name().startswith("chnl_"):
                setattr(self._memobj.bclo, setting.get_name(), setting.value)
            else:
                setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        top = RadioSettings(basic)

        rs = RadioSettingValueInteger(0, 9, _settings.squelch)
        rset = RadioSetting("squelch", "Squelch Level", rs)
        basic.append(rset)

        rs = RadioSettingValueList(TIMEOUTTIMER_LIST,
                                   TIMEOUTTIMER_LIST[_settings.tot - 1])
        rset = RadioSetting("tot", "Time-out timer", rs)
        basic.append(rset)

        rs = RadioSettingValueList(VOICE_LIST,
                                   VOICE_LIST[_settings.voice])
        rset = RadioSetting("voice", "Voice Prompts", rs)
        basic.append(rset)

        rs = RadioSettingValueList(SAVE_LIST,
                                   SAVE_LIST[_settings.savem])
        rset = RadioSetting("savem", "Battery Save Mode", rs)
        basic.append(rset)

        rs = RadioSettingValueList(BACKLIGHT_LIST,
                                   BACKLIGHT_LIST[_settings.backlight])
        rset = RadioSetting("backlight", "Back Light", rs)
        basic.append(rset)

        rs = RadioSettingValueInteger(0, 9, _settings.volume)
        rset = RadioSetting("volume", "Volume", rs)
        basic.append(rset)

        rs = RadioSettingValueInteger(0, 10, _settings.tone)
        rset = RadioSetting("tone", "Tone", rs)
        basic.append(rset)

        rs = RadioSettingValueInteger(1, 22, _settings.chnumber + 1)
        rset = RadioSetting("chnumber", "Channel Number", rs)
        basic.append(rset)

        if self.MODEL == "RB28":
            rs = RadioSettingValueInteger(1, 11, _settings.wxchnumber + 1)
            rset = RadioSetting("wxchnumber", "Weather Channel Number", rs)
            basic.append(rset)

        rs = RadioSettingValueBoolean(_settings.autokeylock)
        rset = RadioSetting("autokeylock", "Auto Key Lock", rs)
        basic.append(rset)

        rs = RadioSettingValueBoolean(_settings.save)
        rset = RadioSetting("save", "Battery Save", rs)
        basic.append(rset)

        rs = RadioSettingValueBoolean(_settings.roger)
        rset = RadioSetting("roger", "End Tone", rs)
        basic.append(rset)

        rs = RadioSettingValueBoolean(_settings.beep)
        rset = RadioSetting("beep", "Beep", rs)
        basic.append(rset)

        rs = RadioSettingValueBoolean(_settings.scan)
        rset = RadioSetting("scan", "Scan", rs)
        basic.append(rset)

        if self.MODEL == "RB28":
            rs = RadioSettingValueBoolean(_settings.wxwarn)
            rset = RadioSetting("wxwarn", "Weather Warn", rs)
            basic.append(rset)

        rs = RadioSettingValueBoolean(_settings.keylock)
        rset = RadioSetting("keylock", "Key Lock", rs)
        basic.append(rset)

        rs = RadioSettingValueList(VOXL_LIST,
                                   VOXL_LIST[_settings.voxl])
        rset = RadioSetting("voxl", "Vox Level", rs)
        basic.append(rset)

        rs = RadioSettingValueList(VOXD_LIST,
                                   VOXD_LIST[_settings.voxd])
        rset = RadioSetting("voxd", "Vox Delay", rs)
        basic.append(rset)

        rs = RadioSettingValueList(self.PFKEY_LIST,
                                   self.PFKEY_LIST[_settings.pfkey_lt])
        rset = RadioSetting("pfkey_lt", "Key Set < Long", rs)
        basic.append(rset)

        rs = RadioSettingValueList(self.PFKEY_LIST,
                                   self.PFKEY_LIST[_settings.pfkey_gt])
        rset = RadioSetting("pfkey_gt", "Key Set > Long", rs)
        basic.append(rset)

        return top

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        bits = element.get_name().split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "chnumber":
                        setattr(obj, setting, int(element.value) - 1)
                    elif setting == "tot":
                        setattr(obj, setting, int(element.value) + 1)
                    elif setting == "wxchnumber":
                        setattr(obj, setting, int(element.value) - 1)
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        # Radios that have always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class RB628Radio(RB28Radio):
    """RETEVIS RB628"""
    VENDOR = "Retevis"
    MODEL = "RB628"

    PFKEY_LIST = PFKEY_EU_LIST
    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=0.50),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _magic = b"PHOGR\x28\x0B"
    _fingerprint = [b"P32073" + b"\x00\xFF", ]
    _upper = 16
    _mem_params = (_upper,  # number of channels
                   )
    _frs = False
    _pmr = True



# Copyright 2018 Rhett Robinson <rrhett@gmail.com> Vers 1.0
# 6/2019: Vers 1.1 Added Settings support, Rick DeWitt <aa0rd@yahoo.com>
# 7/2020: Vers 1.2 Fixed reversed E:* / F:# DTMF codes Issue #8159
# 9/2020: Vers 1.3 Expanded tone mode support and added py3 HAS_FUTURE
#                  bytes() per issue #8233. (icf.py is NOT updated)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from chirp.drivers import icf
from chirp import chirp_common, directory, bitwise
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettingValueInteger, \
    RadioSettingValueFloat, RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
struct {
  u24  freq_flags:6,
       freq:18;
  u16  offset;
  u8   tune_step:4,
       unknown5:2,
       mode:2;
  u8   unknown6:2,
       rtone:6;
  u8   unknown7:2,
       ctone:6;
  u8   unknown8;
  u8   dtcs;
  u8   tmode:4,
       duplex:2,
       dtcs_polarity:2;
  char name[6];
} memory[1002];

#seekto 0x42c0;
u8 used_flags[125];

#seekto 0x433e;
u8 skip_flags[125];
u8 pskip_flags[125];

#seekto 0x4440;
struct {
u8 bank;
u8 index;
} bank_info[1000];

#seekto 0x4c40;
struct {
char com[16];
} comment;

#seekto 0x4c50;
struct {
char name[6];
} bank_names[10];

#seekto 0x4cc8;
struct{
u8 codes[24];
} dtmfcode[16];

#seekto 0x4c8c;
struct {
char nam[6];
} pslnam[10];

#seekto 0x4e80;
struct {
u24  loflags:6,
     lofreq:18;
u24  hiflags:6,
     hifreq:18;
u8  flag:4,
    mode:4;
u8  tstp;
char name[6];
} pgmscanedge[25];

#seekto 0x5000;
struct {
u8  aprichn;
u8  bprichn;
u8  autopwr;
u8  unk5003;
u8  autorptr;
u8  rmtmic;    // 0x05005
u8  pttlock;
u8  bcl;
u8  tot;
u8  actband;
u8  unk500a;
u8  dialspdup;
u8  toneburst;
u8  micgain;
u8  unk500e;
u8  civaddr;
u8  civbaud;
u8  civtcvr;
u8  sqlatt;
u8  sqldly;
u8  unk5014a:4,
    fanspeed:4;
u8  unk5015;
u8  bthvox;
u8  bthvoxlvl;
u8  bthvoxdly;
u8  bthvoxtot;  // 0x05019
u8  btoothon;
u8  btoothauto;
u8  bthdset;
u8  bthhdpsav;
u8  bth1ptt;    // 0x0501e
u8  bthpttbeep;
u8  bthcustbeep;    // 0x05020
u8  ascanpause;
u8  bscanpause;
u8  ascanresume;
u8  bscanresume;
u8  dtmfspd;
u8  unk5026;
u8  awxalert;
u8  bwxalert;
u8  aprgskpscn;
u8  bprgskpscn;
u8  memname;
u8  contrast;
u8  autodimtot;
u8  autodim;
u8  unk502f;
u8  backlight;
u8  unk5031;
u8  unk5032;
u8  openmsg;
u8  beeplvl;    // 0x05034
u8  keybeep;
u8  scanstpbeep;
u8  bandedgbeep;
u8  subandmute;
u8  atmpskiptym;
u8  btmpskiptym;
u32 vfohome;
u8  vfohomeset;
u16 homech;
u8  mickyrxf1;
u8  mickyrxf2;
u8  mickyrxup;
u8  mickyrxdn;  // 0x05045
u8  bthplaykey;
u8  bthfwdkey;
u8  bthrwdkey;
u8  mickytxf1;
u8  mickytxf2;
u8  mickytxup;
u8  mickytxdn;
u8  unk504d;
u8  unk504e;
u8  unk504f;
u8  homebeep;
u8  bthfctn;
u8  unk5052;
u8  unk5053;
u8  ifxchg;
u8  airbandch;
u8  vhfpower;
u8  uhfpower;
u8  unk5058;
u8  unk5059;
u8  unk505a;
u8  unk505b;
u8  unk505c;
u8  unk505d;
u8  unk505e:6,
    rpthangup:1,
    unk505e2:1;
u8  unk505f;
} settings;

#seekto 0x5220;
struct {
u16  left_memory;
u16  right_memory;
} initmem;

#seekto 0x523e;
struct {
u8  awxchan;
u8  bwxchan;
} abwx;

#seekto 0x5250;
struct {
u8  alnk[2];
u16 unk5252;
u8  blnk[2];
} banklink;

#seekto 0x5258;
struct {
u8  msk[4];
} pslgrps[25];

#seekto 0x5280;
u16  mem_writes_count;
"""

MDUPLEX = ["", "-", "+"]
MSKIPS = ["", "S", "P"]
MMODES = ["FM", "NFM", "AM", "NAM"]
MTONE_MODES = chirp_common.TONE_MODES
MTMODES = ["", "Tone", "TonePoc", "TSQL", "DTCSPoc", "DTCS",
           "TSQL-R", "DTCS-R", "DTC.OFF", "TON.DTC",
           "DTC.TSQ", "TON.TSQ"]
MCROSS_MODES = ["Tone->Tone", "DTCS->", "Tone->DTCS", "DTCS->Tone"]
MDTCSP = ["NN", "NR", "RN", "RR"]
MDTMF_CHARS = list("0123456789ABCD*#")
MBANKLINK_CHARS = list("ABCDEFGHIJ")
AUTOREPEATER = ["OFF", "DUP", "DUP.TON"]
MICKEYOPTS = ["Off", "Up", "Down", "Vol Up", "Vol Down", "SQL Up",
              "SQL Down", "Monitor", "Call", "MR (Ch 0)", "MR (Ch 1)",
              "VFO/MR", "Home Chan", "Band/Bank", "Scan", "Temp Skip",
              "Main", "Mode", "Low", "Dup", "Priority", "Tone", "MW", "Mute",
              "T-Call", "DTMF Direct"]


class IC2730Bank(icf.IcomNamedBank):
    """An IC2730 bank"""
    def get_name(self):
        _banks = self._model._radio._memobj.bank_names
        return str(_banks[self.index].name).rstrip()

    def set_name(self, name):
        _banks = self._model._radio._memobj.bank_names
        _banks[self.index].name = str(name).ljust(6)[:6]


def _get_special():
    special = {"C0": -2, "C1": -1}
    return special


def _resolve_memory_number(number):
    if isinstance(number, str):
        return _get_special()[number]
    else:
        return number


def _wipe_memory(mem, char):
    mem.set_raw(char * (mem.size() // 8))


@directory.register
class IC2730Radio(icf.IcomCloneModeRadio):
    """Icom IC-2730A"""
    VENDOR = "Icom"
    MODEL = "IC-2730A"

    _model = "\x35\x98\x00\x01"
    _memsize = 21312  # 0x5340
    _endframe = "Icom Inc\x2e4E"

    _ranges = [(0x0000, 0x5300, 64),
               (0x5300, 0x5310, 16),
               (0x5310, 0x5340, 48)]

    _num_banks = 10
    _bank_class = IC2730Bank
    _can_hispeed = True
    _raw_frames = True
    _highbit_flip = True

    _icf_data = {
        'MapRev': 1,
        'EtcData': 0,  # This might be wrong
        'Comment': '',
        'recordsize': 32,
    }

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.info = ('Click the Special Channels tab on the main screen to '
                   'access the C0 and C1 frequencies.\n'
                   'The Pocket Beep tone/dtcs modes are not supported by '
                   'CHIRP. Set those manually.\n')

        rp.pre_download = _(
            "Follow these instructions to download your config:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable to the Speaker-2 jack\n"
            "3 - Turn on your radio\n"
            "4 - Radio > Download from radio\n"
            "5 - Disconnect the interface cable! Otherwise there will be\n"
            "    no right-side audio!\n")
        rp.pre_upload = _(
            "Follow these instructions to upload your config:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable to the Speaker-2 jack\n"
            "3 - Turn on your radio\n"
            "4 - Radio > Upload to radio\n"
            "5 - Disconnect the interface cable, otherwise there will be\n"
            "    no right-side audio!\n"
            "6 - Cycle power on the radio to exit clone mode\n")
        return rp

    def _get_bank(self, loc):
        _bank = self._memobj.bank_info[loc]
        _bank.bank = _bank.bank & 0x1F      # Bad index filter, fix issue #7031
        if _bank.bank == 0x1F:
            return None
        else:
            return _bank.bank

    def _set_bank(self, loc, bank):
        _bank = self._memobj.bank_info[loc]
        if bank is None:
            _bank.bank = 0x1F
        else:
            _bank.bank = bank

    def _get_bank_index(self, loc):
        _bank = self._memobj.bank_info[loc]
        return _bank.index

    def _set_bank_index(self, loc, index):
        _bank = self._memobj.bank_info[loc]
        _bank.index = index

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank_index = True
        rf.has_bank_names = True
        rf.has_cross = True
        rf.has_ctone = True
        rf.has_dtcs = True
        rf.has_rx_dtcs = False
        rf.has_dtcs_polarity = True
        rf.requires_call_lists = False
        rf.memory_bounds = (0, 999)
        rf.valid_modes = MMODES
        rf.valid_tmodes = MTONE_MODES
        rf.valid_duplexes = MDUPLEX
        rf.valid_tuning_steps = list(chirp_common.TUNING_STEPS[0:9])
        rf.valid_bands = [(118000000, 174000000),
                          (375000000, 550000000)]
        rf.valid_skips = MSKIPS
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.valid_cross_modes = MCROSS_MODES
        rf.valid_name_length = 6
        rf.valid_special_chans = sorted(_get_special().keys())

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_memory(self, number):
        mem = chirp_common.Memory()
        mem.number = _resolve_memory_number(number)
        if mem.number >= 0:
            _mem = self._memobj.memory[mem.number]
            bitpos = (1 << (number % 8))
            bytepos = number / 8
            _used = self._memobj.used_flags[bytepos]
            is_used = ((_used & bitpos) == 0)

            _skip = self._memobj.skip_flags[bytepos]
            _pskip = self._memobj.pskip_flags[bytepos]
            if _skip & bitpos:
                mem.skip = MSKIPS[1]     # "S"
            elif _pskip & bitpos:
                mem.skip = MSKIPS[2]    # "P"
            if not is_used:
                mem.empty = True
                return mem
        else:   # C0, C1 specials
            _mem = self._memobj.memory[1002 + mem.number]

        # _mem.freq is stored as a multiple of a tuning step
        frequency_flags = int(_mem.freq_flags)
        frequency_multiplier = 5000
        offset_multiplier = 5000
        if frequency_flags & 0x08:
            frequency_multiplier = 6250
        if frequency_flags & 0x01:
            offset_multiplier = 6250
        if frequency_flags & 0x10:
            frequency_multiplier = 8333
        if frequency_flags & 0x02:
            offset_multiplier = 8333

        if frequency_flags & 0x10:  # fix underflow
            val = int(_mem.freq) * frequency_multiplier
            mem.freq = round(val)
        else:
            mem.freq = int(_mem.freq) * frequency_multiplier
        if frequency_flags & 0x02:
            val = int(_mem.offset) * offset_multiplier
            mem.offset = round(val)
        else:
            mem.offset = int(_mem.offset) * offset_multiplier
        mem.rtone = chirp_common.TONES[_mem.rtone]
        mem.ctone = chirp_common.TONES[_mem.ctone]
        tmx = 0     # MTONE_MODES index default
        cmx = 0     # MCROSS_MODES index
        if _mem.tmode == 1:     # Tone
            tmx = 1
        elif _mem.tmode == 3:   # TSQL
            tmx = 2
        elif _mem.tmode == 5:   # DTCS
            tmx = 3
        elif _mem.tmode == 6:   # TSQL-R
            tmx = 5
        elif _mem.tmode == 7:   # DTCS-R
            tmx = 4
        elif _mem.tmode == 8:   # DTC.OFF
            tmx = 6             # Real Cross modes now
            cmx = 1
        elif _mem.tmode == 9:   # TON.DTC
            tmx = 6
            cmx = 2
        elif _mem.tmode == 10:   # DTC.TSQ
            tmx = 6
            cmx = 3
        elif _mem.tmode == 11:  # TON.TSQ
            tmx = 6
            cmx = 0
        mem.tmode = MTONE_MODES[tmx]
        mem.cross_mode = MCROSS_MODES[cmx]
        mem.duplex = MDUPLEX[_mem.duplex]
        mem.mode = MMODES[_mem.mode]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.dtcs_polarity = MDTCSP[_mem.dtcs_polarity]

        if _mem.tune_step > 8:
            mem.tuning_step = 5.0  # Sometimes TS is garbage?
        else:
            mem.tuning_step = chirp_common.TUNING_STEPS[_mem.tune_step]
        mem.name = str(_mem.name).rstrip()
        if mem.number == -2:
            mem.name = "C0"
            mem.extd_number = "C0"
        if mem.number == -1:
            mem.name = "C1"
            mem.extd_number = "C1"

        return mem

    def set_memory(self, mem):
        if mem.number >= 0:       # Normal
            bitpos = (1 << (mem.number % 8))
            bytepos = mem.number / 8

            _mem = self._memobj.memory[mem.number]
            _used = self._memobj.used_flags[bytepos]

            was_empty = _used & bitpos

            skip = self._memobj.skip_flags[bytepos]
            pskip = self._memobj.pskip_flags[bytepos]
            if mem.skip == "S":
                skip |= bitpos
            else:
                skip &= ~bitpos
            if mem.skip == "P":
                pskip |= bitpos
            else:
                pskip &= ~bitpos

            if mem.empty:
                _used |= bitpos
                _wipe_memory(_mem, "\xFF")
                self._set_bank(mem.number, None)
                return

            _used &= ~bitpos
            if was_empty:
                _wipe_memory(_mem, "\x00")
            _mem.name = mem.name.ljust(6)
        else:       # Specials: -2 and -1
            _mem = self._memobj.memory[1002 + mem.number]

        # Common to both types
        frequency_flags = 0x00
        frequency_multiplier = 5000
        offset_multiplier = 5000
        if mem.freq % 5000 != 0 and mem.freq % 6250 == 0:
            frequency_flags |= 0x08
            frequency_multiplier = 6250
        elif mem.freq % 8333 == 0:
            frequency_flags |= 0x10
            frequency_multiplier = 8333
        if mem.offset % 5000 != 0 and mem.offset % 6250 == 0:
            frequency_flags |= 0x01
            offset_multiplier = 6250
        elif mem.offset % 8333 == 0:
            frequency_flags |= 0x02
            offset_multiplier = 8333
        _mem.freq = mem.freq / frequency_multiplier
        _mem.offset = mem.offset / offset_multiplier
        _mem.freq_flags = frequency_flags
        _mem.rtone = chirp_common.TONES.index(mem.rtone)
        _mem.ctone = chirp_common.TONES.index(mem.ctone)
        tmx = MTONE_MODES.index(mem.tmode)
        cmx = MCROSS_MODES.index(mem.cross_mode)
        if tmx != 6:        # Not Cross
            if tmx < 2:     # "", Tone
                _mem.tmode = tmx
            if tmx == 2:    # TSQL
                _mem.tmode = 3
            if tmx == 3:    # DTCS
                _mem.tmode = 5
            if tmx == 4:    # DTCS-R
                _mem.tmode = 7
            if tmx == 5:    # TSQL-R
                _mem.tmode = 6
        else:           # Cross modes
            if cmx == 0:    # Tone->Tone
                _mem.tmode = 11
            if cmx == 1:    # DTCS->
                _mem.tmode = 8
            if cmx == 2:    # Tone->DTCS
                _mem.tmode = 9
            if cmx == 3:    # DTCS->Tone
                _mem.tmode = 10
        _mem.duplex = MDUPLEX.index(mem.duplex)
        _mem.mode = MMODES.index(mem.mode)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.dtcs_polarity = MDTCSP.index(mem.dtcs_polarity)
        _mem.tune_step = chirp_common.TUNING_STEPS.index(mem.tuning_step)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def get_settings(self):
        """Translate the MEM_FORMAT structs into settings in the UI"""
        # Define mem struct write-back shortcuts
        _sets = self._memobj.settings
        _cmnt = self._memobj.comment
        _wxch = self._memobj.abwx
        _dtm = self._memobj.dtmfcode
        _pses = self._memobj.pgmscanedge
        _bklk = self._memobj.banklink

        basic = RadioSettingGroup("basic", "Basic Settings")
        mickey = RadioSettingGroup("mickey", "Microphone Keys")
        bluet = RadioSettingGroup("bluet", "Bluetooth")
        disp = RadioSettingGroup("disp", "Display")
        sound = RadioSettingGroup("sound", "Sounds")
        dtmf = RadioSettingGroup("dtmf", "DTMF Codes")
        abset = RadioSettingGroup("abset", "A/B Band Settings")
        edges = RadioSettingGroup("edges", "Program Scan Edges")
        pslnk = RadioSettingGroup("pslnk", "Program Scan Links")
        other = RadioSettingGroup("other", "Other Settings")

        group = RadioSettings(basic, disp, sound, mickey, dtmf,
                              abset, bluet, edges, pslnk, other)

        def mic_keys(setting, obj, atrb):
            """ Callback to set subset of mic key options """
            stx = str(setting.value)
            value = MICKEYOPTS.index(stx)
            setattr(obj, atrb, value)
            return

        def hex_val(setting, obj, atrb):
            """ Callback to store string as hex values """
            value = int(str(setting.value), 16)
            setattr(obj, atrb, value)
            return

        def unpack_str(codestr):
            """Convert u8 DTMF array to a string: NOT a callback."""
            stx = ""
            for i in range(0, 24):    # unpack up to ff
                if codestr[i] != 0xff:      # Issue  8159 fix
                    if codestr[i] == 0x0E:
                        stx += "*"
                    elif codestr[i] == 0x0F:
                        stx += "#"
                    else:
                        stx += format(int(codestr[i]), '0X')
            return stx

        def pack_chars(setting, obj, atrb, ndx):
            """Callback to build 0-9,A-D,*# nibble array from string"""
            # String will be f padded to 24 bytes
            # Chars are stored as hex values
            ary = []
            stx = str(setting.value).upper()
            stx = stx.strip()       # trim spaces
            # Remove illegal characters first
            sty = ""
            for j in range(0, len(stx)):
                if stx[j] in MDTMF_CHARS:
                    sty += stx[j]
            for j in range(0, 24):
                if j < len(sty):        # Issue 8159 fix
                    if sty[j] == "*":
                        chrv = 0xE
                    elif sty[j] == "#":
                        chrv = 0xF
                    else:
                        chrv = int(sty[j], 16)
                else:   # pad to 24 bytes
                    chrv = 0xFF
                ary.append(chrv)    # append byte
            setattr(obj[ndx], atrb, ary)
            return

        def myset_comment(setting, obj, atrb, knt):
            """ Callback to create space-padded char array"""
            stx = str(setting.value)
            for i in range(0, knt):
                if i > len(stx):
                    str.append(0x20)
            setattr(obj, atrb, stx)
            return

        def myset_psnam(setting, obj, ndx, atrb, knt):
            """ Callback to generate space-padded, uppercase char array """
            # This sub also is specific to object arrays
            stx = str(setting.value).upper()
            for i in range(0, knt):
                if i > len(stx):
                    str.append(0x20)
            setattr(obj[ndx], atrb, stx)
            return

        def myset_frqflgs(setting, obj, ndx, flg, frq):
            """ Callback to gen flag/freq pairs """
            vfrq = float(str(setting.value))
            vfrq = int(vfrq * 1000000)
            vflg = 0x10
            if vfrq % 6250 == 0:
                vflg = 0x08
                vfrq = int(vfrq / 6250)
            elif vfrq % 5000 == 0:
                vflg = 0
                vfrq = int(vfrq / 5000)
            else:
                vfrq = int(vfrq / 8333)
            setattr(obj[ndx], flg, vflg)
            setattr(obj[ndx], frq, vfrq)
            return

        def banklink(ary):
            """ Sub to generate A-J string from 2-byte bit pattern """
            stx = ""
            for kx in range(0, 10):
                if kx < 8:
                    val = ary[0]
                    msk = 1 << kx
                else:
                    val = ary[1]
                    msk = 1 << (kx - 8)
                if val & msk:
                    stx += chr(kx + 65)
                else:
                    stx += "_"
            return stx

        def myset_banklink(setting, obj, atrb):
            """Callback to create 10-bit, u8[2] array from 10 char string"""
            stx = str(setting.value).upper()
            ary = [0, 0]
            for kx in range(0, 10):
                if stx[kx] == chr(kx + 65):
                    if kx < 8:
                        ary[0] = ary[0] + (1 << kx)
                    else:
                        ary[1] = ary[1] + (1 << (kx - 8))
            setattr(obj, atrb, ary)
            return

        def myset_tsopt(setting, obj, ndx, atrb, bx):
            """ Callback to set scan Edge tstep """
            stx = str(setting.value)
            flg = 0
            if stx == "-":
                val = 0xff
            else:
                if bx == 1:  # Air band
                    if stx == "Auto":
                        val = 0xe
                    elif stx == "25k":
                        val = 8
                    elif stx == "8.33k":
                        val = 2
                else:       # VHF or UHF
                    optx = ["-", "5k", "6.25k", "10k", "12.5k", "15k",
                            "20k", "25k", "30k", "50k"]
                    val = optx.index(stx) + 1
            setattr(obj[ndx], atrb, val)
            # and set flag
            setattr(obj[ndx], "flag", flg)
            return

        def myset_mdopt(setting, obj, ndx, atrb, bx):
            """ Callback to set Scan Edge mode """
            stx = str(setting.value)
            if stx == "-":
                val = 0xf
            elif stx == "FM":
                val = 0
            else:
                val = 1
            setattr(obj[ndx], atrb, val)
            return

        def myset_bitmask(setting, obj, ndx, atrb, knt):
            """ Callback to generate byte-array bitmask from string"""
            # knt is BIT count to process
            lsx = str(setting.value).strip().split(",")
            for kx in range(0, len(lsx)):
                try:
                    lsx[kx] = int(lsx[kx])
                except Exception:
                    lsx[kx] = -99   # will nop
            ary = [0, 0, 0, 0xfe]
            for kx in range(0, knt):
                if kx < 8:
                    if kx in lsx:
                        ary[0] += 1 << kx
                elif kx >= 8 and kx < 16:
                    if kx in lsx:
                        ary[1] += 1 << (kx - 8)
                elif kx >= 16 and kx < 24:
                    if kx in lsx:
                        ary[2] += 1 << (kx - 16)
                else:
                    if kx in lsx:   # only bit 25
                        ary[3] += 1
            setattr(obj[ndx], atrb, ary)
            return

        # --- Basic
        options = ["Off", "S-Meter Squelch", "ATT"]
        rx = RadioSettingValueList(options, options[_sets.sqlatt])
        rset = RadioSetting("settings.sqlatt", "Squelch/ATT", rx)
        basic.append(rset)

        options = ["Short", "Long"]
        rx = RadioSettingValueList(options, options[_sets.sqldly])
        rset = RadioSetting("settings.sqldly", "Squelch Delay", rx)
        basic.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.pttlock))
        rset = RadioSetting("settings.pttlock", "PTT Lockout", rx)
        basic.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bcl))
        rset = RadioSetting("settings.bcl", "Busy Channel Lockout", rx)
        basic.append(rset)

        options = ["Off", "1m", "3m", "5m", "10m", "15m", "30m"]
        rx = RadioSettingValueList(options, options[_sets.tot])
        rset = RadioSetting("settings.tot", "Tx Timeout", rx)
        basic.append(rset)

        val = int(_sets.vfohome)
        if val == 0xffff:
            val = 0
        val = val / 1000000.0
        rx = RadioSettingValueFloat(0.0, 550.0, val, 0.005, 4)
        rx.set_mutable(False)
        rset = RadioSetting("settings.vfohome", "Home VFO (Read-Only)", rx)
        basic.append(rset)

        val = _sets.homech
        if val == 0xffff:
            val = -1
        rx = RadioSettingValueInteger(-1, 999, val)
        rx.set_mutable(False)
        rset = RadioSetting("settings.homech",
                            "Home Channel (Read-Only)", rx)
        basic.append(rset)

        options = ["1", "2", "3", "4"]
        rx = RadioSettingValueList(options, options[_sets.micgain])
        rset = RadioSetting("settings.micgain", "Microphone Gain", rx)
        basic.append(rset)

        _bmem = self._memobj.initmem
        rx = RadioSettingValueInteger(0, 999, _bmem.left_memory)
        rset = RadioSetting("initmem.left_memory",
                            "Left Bank Initial Mem Chan", rx)
        basic.append(rset)

        rx = RadioSettingValueInteger(0, 999, _bmem.right_memory)
        rset = RadioSetting("initmem.right_memory",
                            "Right Bank Initial Mem Chan", rx)
        basic.append(rset)

        stx = ""
        for i in range(0, 16):
            stx += chr(int(_cmnt.com[i]))
        stx = stx.rstrip()
        rx = RadioSettingValueString(0, 16, stx)
        rset = RadioSetting("comment.com", "Comment (16 chars)", rx)
        rset.set_apply_callback(myset_comment, _cmnt, "com", 16)
        basic.append(rset)

        # --- Other
        rset = RadioSetting("drv_clone_speed", "Use Hi-Speed Clone",
                            RadioSettingValueBoolean(self._can_hispeed))
        other.append(rset)

        options = ["Single", "All", "Ham"]
        rx = RadioSettingValueList(options, options[_sets.actband])
        rset = RadioSetting("settings.actband", "Active Band", rx)
        other.append(rset)

        options = ["Slow", "Mid", "Fast", "Auto"]
        rx = RadioSettingValueList(options, options[_sets.fanspeed])
        rset = RadioSetting("settings.fanspeed", "Fan Speed", rx)
        other.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.dialspdup))
        rset = RadioSetting("settings.dialspdup", "Dial Speed-Up", rx)
        other.append(rset)

        options = ["Off", "On(Dup)", "On(Dup+Tone)"]
        rx = RadioSettingValueList(options, options[_sets.autorptr])
        rset = RadioSetting("settings.autorptr", "Auto Repeater", rx)
        other.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.rmtmic))
        rset = RadioSetting("settings.rmtmic",
                            "One-Touch PTT (Remote Mic)", rx)
        other.append(rset)

        options = ["Low", "Mid", "High"]
        rx = RadioSettingValueList(options, options[_sets.vhfpower])
        rset = RadioSetting("settings.vhfpower", "VHF Power Default", rx)
        other.append(rset)

        rx = RadioSettingValueList(options, options[_sets.uhfpower])
        rset = RadioSetting("settings.uhfpower", "UHF Power Default", rx)
        other.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.toneburst))
        rset = RadioSetting("settings.toneburst", "1750 Hz Tone Burst", rx)
        other.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.ifxchg))
        rset = RadioSetting("settings.ifxchg", "IF Exchange", rx)
        other.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.rpthangup))
        rset = RadioSetting("settings.rpthangup",
                            "Repeater Hang up Timeout", rx)
        other.append(rset)

        stx = str(_sets.civaddr)[2:]    # Hex value
        rx = RadioSettingValueString(1, 2, stx)
        rset = RadioSetting("settings.civaddr", "CI-V Address (90)", rx)
        rset.set_apply_callback(hex_val, _sets, "civaddr")
        other.append(rset)

        options = ["1200", "2400", "4800", "9600", "19200", "Auto"]
        rx = RadioSettingValueList(options, options[_sets.civbaud])
        rset = RadioSetting("settings.civbaud", "CI-V Baud Rate (bps)", rx)
        other.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.civtcvr))
        rset = RadioSetting("settings.civtcvr", "CI-V Transceive", rx)
        other.append(rset)

        # A/B Band Settings
        options = ["Off", "On", "Bell"]
        rx = RadioSettingValueList(options, options[_sets.aprichn])
        rset = RadioSetting("settings.aprichn",
                            "A Band: VFO Priority Watch Mode", rx)
        abset.append(rset)

        options = ["2", "4", "6", "8", "10", "12", "14",
                   "16", "18", "20", "Hold"]
        rx = RadioSettingValueList(options, options[_sets.ascanpause])
        rset = RadioSetting("settings.ascanpause",
                            "-- A Band: Scan Pause Time (Secs)", rx)
        abset.append(rset)

        options = ["0", "1", "2", "3", "4", "5", "Hold"]
        rx = RadioSettingValueList(options, options[_sets.ascanresume])
        rset = RadioSetting("settings.ascanresume",
                            "-- A Band: Scan Resume Time (Secs)", rx)
        abset.append(rset)

        options = ["5", "10", "15"]
        rx = RadioSettingValueList(options, options[_sets.atmpskiptym])
        rset = RadioSetting("settings.atmpskiptym",
                            "-- A Band: Temp Skip Time (Secs)", rx)
        abset.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.aprgskpscn))
        rset = RadioSetting("settings.aprgskpscn",
                            "-- A Band: Program Skip Scan", rx)
        abset.append(rset)

        rx = RadioSettingValueString(10, 10, banklink(_bklk.alnk))
        rset = RadioSetting("banklink.alnk",
                            "-- A Band Banklink (use _ to skip)", rx)
        rset.set_apply_callback(myset_banklink, _bklk, "alnk")
        abset.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.awxalert))
        rset = RadioSetting("settings.awxalert",
                            "-- A Band: Weather Alert", rx)
        abset.append(rset)

        # Use list for Wx chans since chan 1 = index 0
        options = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
        rx = RadioSettingValueList(options, options[_wxch.awxchan])
        rset = RadioSetting("abwx.awxchan", "-- A Band: Weather Channel", rx)
        abset.append(rset)

        options = ["Off", "On", "Bell"]
        rx = RadioSettingValueList(options, options[_sets.bprichn])
        rset = RadioSetting("settings.bprichn",
                            "B Band: VFO Priority Watch Mode", rx)
        abset.append(rset)

        options = ["2", "4", "6", "8", "10", "12", "14",
                   "16", "18", "20", "Hold"]
        rx = RadioSettingValueList(options, options[_sets.bscanpause])
        rset = RadioSetting("settings.bscanpause",
                            "-- B Band: Scan Pause Time (Secs)", rx)
        abset.append(rset)

        options = ["0", "1", "2", "3", "4", "5", "Hold"]
        rx = RadioSettingValueList(options, options[_sets.bscanresume])
        rset = RadioSetting("settings.bscanresume",
                            "-- B Band: Scan Resume Time (Secs)", rx)
        abset.append(rset)

        options = ["5", "10", "15"]
        rx = RadioSettingValueList(options, options[_sets.btmpskiptym])
        rset = RadioSetting("settings.btmpskiptym",
                            "-- B Band: Temp Skip Time (Secs)", rx)
        abset.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bprgskpscn))
        rset = RadioSetting("settings.bprgskpscn",
                            "-- B Band: Program Skip Scan", rx)
        abset.append(rset)

        rx = RadioSettingValueString(10, 10, banklink(_bklk.blnk))
        rset = RadioSetting("banklink.blnk",
                            "-- B Band Banklink (use _ to skip)", rx)
        rset.set_apply_callback(myset_banklink, _bklk, "blnk")
        abset.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bwxalert))
        rset = RadioSetting("settings.bwxalert",
                            "-- B Band: Weather Alert", rx)
        abset.append(rset)

        options = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
        rx = RadioSettingValueList(options, options[_wxch.bwxchan])
        rset = RadioSetting("abwx.bwxchan", "-- B Band: Weather Channel", rx)
        abset.append(rset)

        # --- Microphone Keys
        # The Mic keys get weird: stored values are indecis to the full
        # options list, but only a subset is valid...
        shortopts = ["Off", "Monitor", "MR (Ch 0)", "MR (Ch 1)", "Band/Bank",
                     "Scan", "Temp Skip", "Mode", "Low", "Dup", "Priority",
                     "Tone", "MW", "Mute", "DTMF Direct", "T-Call"]
        ptr = shortopts.index(MICKEYOPTS[_sets.mickyrxf1])
        rx = RadioSettingValueList(shortopts, shortopts[ptr])
        rset = RadioSetting("settings.mickyrxf1",
                            "During Rx/Standby [F-1]", rx)
        rset.set_apply_callback(mic_keys, _sets, "mickyrxf1")
        mickey.append(rset)

        ptr = shortopts.index(MICKEYOPTS[_sets.mickyrxf2])
        rx = RadioSettingValueList(shortopts, shortopts[ptr])
        rset = RadioSetting("settings.mickyrxf2",
                            "During Rx/Standby [F-2]", rx)
        rset.set_apply_callback(mic_keys, _sets, "mickyrxf2")
        mickey.append(rset)

        options = ["Off", "Low", "T-Call"]      # NOT a subset of MICKEYOPTS
        rx = RadioSettingValueList(options, options[_sets.mickytxf1])
        rset = RadioSetting("settings.mickytxf1", "During Tx [F-1]", rx)
        mickey.append(rset)

        rx = RadioSettingValueList(options, options[_sets.mickytxf2])
        rset = RadioSetting("settings.mickytxf2", "During Tx [F-2]", rx)
        mickey.append(rset)

        # These next two get the full options list
        rx = RadioSettingValueList(MICKEYOPTS, MICKEYOPTS[_sets.mickyrxup])
        rset = RadioSetting("settings.mickyrxup",
                            "During Rx/Standby [Up]", rx)
        mickey.append(rset)

        rx = RadioSettingValueList(MICKEYOPTS, MICKEYOPTS[_sets.mickyrxdn])
        rset = RadioSetting("settings.mickyrxdn",
                            "During Rx/Standby [Down]", rx)
        mickey.append(rset)

        options = ["Off", "Low", "T-Call"]
        rx = RadioSettingValueList(options, options[_sets.mickytxup])
        rset = RadioSetting("settings.mickytxup", "During Tx [Up]", rx)
        mickey.append(rset)

        rx = RadioSettingValueList(options, options[_sets.mickytxdn])
        rset = RadioSetting("settings.mickytxdn", "During Tx [Down]", rx)
        mickey.append(rset)

        # --- Bluetooth
        rx = RadioSettingValueBoolean(bool(_sets.btoothon))
        rset = RadioSetting("settings.btoothon", "Bluetooth", rx)
        bluet.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.btoothauto))
        rset = RadioSetting("settings.btoothauto", "Auto Connect", rx)
        bluet.append(rset)

        options = ["Headset Only", "Headset & Speaker"]
        rx = RadioSettingValueList(options, options[_sets.bthdset])
        rset = RadioSetting("settings.bthdset", "Headset Audio", rx)
        bluet.append(rset)

        options = ["Normal", "Microphone", "PTT (Audio:Main)",
                   "PTT(Audio:Controller)"]
        rx = RadioSettingValueList(options, options[_sets.bthfctn])
        rset = RadioSetting("settings.bthfctn", "Headset Function", rx)
        bluet.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bthvox))
        rset = RadioSetting("settings.bthvox", "Vox", rx)
        bluet.append(rset)

        options = ["Off", "1.0", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
        rx = RadioSettingValueList(options, options[_sets.bthvoxlvl])
        rset = RadioSetting("settings.bthvoxlvl", "Vox Level", rx)
        bluet.append(rset)

        options = ["0.5", "1.0", "1.5", "2.0", "2.5", "3.0"]
        rx = RadioSettingValueList(options, options[_sets.bthvoxdly])
        rset = RadioSetting("settings.bthvoxdly", "Vox Delay (Secs)", rx)
        bluet.append(rset)

        options = ["Off", "1", "2", "3", "4", "5", "10", "15"]
        rx = RadioSettingValueList(options, options[_sets.bthvoxtot])
        rset = RadioSetting("settings.bthvoxtot", "Vox Time-Out (Mins)", rx)
        bluet.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bthhdpsav))
        rset = RadioSetting("settings.bthhdpsav",
                            "ICOM Headset Power-Save", rx)
        bluet.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bth1ptt))
        rset = RadioSetting("settings.bth1ptt",
                            "ICOM Headset One-Touch PTT", rx)
        bluet.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bthpttbeep))
        rset = RadioSetting("settings.bthpttbeep",
                            "ICOM Headset PTT Beep", rx)
        bluet.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bthcustbeep))
        rset = RadioSetting("settings.bthcustbeep",
                            "ICOM Headset Custom Key Beep", rx)
        bluet.append(rset)

        rx = RadioSettingValueList(MICKEYOPTS, MICKEYOPTS[_sets.bthplaykey])
        rset = RadioSetting("settings.bthplaykey",
                            "ICOM Headset Custom Key [Play]", rx)
        bluet.append(rset)

        rx = RadioSettingValueList(MICKEYOPTS, MICKEYOPTS[_sets.bthfwdkey])
        rset = RadioSetting("settings.bthfwdkey",
                            "ICOM Headset Custom Key [Fwd]", rx)
        bluet.append(rset)

        rx = RadioSettingValueList(MICKEYOPTS, MICKEYOPTS[_sets.bthrwdkey])
        rset = RadioSetting("settings.bthrwdkey",
                            "ICOM Headset Custom Key [Rwd]", rx)
        bluet.append(rset)

        # ---- Display
        options = ["1: Dark", "2", "3", "4: Bright"]
        rx = RadioSettingValueList(options, options[_sets.backlight])
        rset = RadioSetting("settings.backlight", "Backlight Level", rx)
        disp.append(rset)

        options = ["Off", "Auto-Off", "Auto-1", "Auto-2", "Auto-3"]
        rx = RadioSettingValueList(options, options[_sets.autodim])
        rset = RadioSetting("settings.autodim", "Auto Dimmer", rx)
        disp.append(rset)

        options = ["5", "10"]
        rx = RadioSettingValueList(options, options[_sets.autodimtot])
        rset = RadioSetting("settings.autodimtot",
                            "Auto-Dimmer Timeout (Secs)", rx)
        disp.append(rset)

        options = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
        rx = RadioSettingValueList(options, options[_sets.contrast])
        rset = RadioSetting("settings.contrast", "LCD Contrast", rx)
        disp.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.openmsg))
        rset = RadioSetting("settings.openmsg", "Opening Message", rx)
        disp.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.memname))
        rset = RadioSetting("settings.memname", "Memory Names", rx)
        disp.append(rset)

        options = ["CH ID", "Frequency"]
        rx = RadioSettingValueList(options, options[_sets.airbandch])
        rset = RadioSetting("settings.airbandch", "Air Band Display", rx)
        disp.append(rset)

        # -- Sounds
        rx = RadioSettingValueInteger(0, 9, _sets.beeplvl)
        rset = RadioSetting("settings.beeplvl", "Beep Level", rx)
        sound.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.homebeep))
        rset = RadioSetting("settings.homebeep", "Home Chan Beep", rx)
        sound.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.keybeep))
        rset = RadioSetting("settings.keybeep", "Key Touch Beep", rx)
        sound.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.bandedgbeep))
        rset = RadioSetting("settings.bandedgbeep", "Band Edge Beep", rx)
        sound.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.scanstpbeep))
        rset = RadioSetting("settings.scanstpbeep", "Scan Stop Beep", rx)
        sound.append(rset)

        options = ["Off", "Mute", "Beep", "Mute & Beep"]
        rx = RadioSettingValueList(options, options[_sets.subandmute])
        rset = RadioSetting("settings.subandmute", "Sub Band Mute", rx)
        sound.append(rset)

        # --- DTMF Codes
        options = ["100", "200", "300", "500"]
        rx = RadioSettingValueList(options, options[_sets.dtmfspd])
        rset = RadioSetting("settings.dtmfspd", "DTMF Speed (mSecs)", rx)
        dtmf.append(rset)

        for kx in range(0, 16):
            stx = unpack_str(_dtm[kx].codes)
            rx = RadioSettingValueString(0, 24, stx)
            # NOTE the / to indicate indexed array
            rset = RadioSetting("dtmfcode/%d.codes" % kx,
                                "DTMF Code %X" % kx, rx)
            rset.set_apply_callback(pack_chars, _dtm, "codes", kx)
            dtmf.append(rset)

        # --- Program Scan Edges
        for kx in range(0, 25):
            stx = ""
            for i in range(0, 6):
                stx += chr(int(_pses[kx].name[i]))
            stx = stx.rstrip()
            rx = RadioSettingValueString(0, 6, stx)
            rset = RadioSetting("pgmscanedge/%d.name" % kx,
                                "Program Scan %d Name" % kx, rx)
            rset.set_apply_callback(myset_psnam, _pses, kx, "name", 6)
            edges.append(rset)

            # Freq's use the multiplier flags
            fmult = 5000.0
            if _pses[kx].loflags == 0x10:
                fmult = 8333
            if _pses[kx].loflags == 0x08:
                fmult = 6250.0
            flow = (int(_pses[kx].lofreq) * fmult) / 1000000.0
            flow = round(flow, 4)

            fmult = 5000.0
            if _pses[kx].hiflags == 0x10:
                fmult = 8333
            if _pses[kx].hiflags == 0x08:
                fmult = 6250.0
            fhigh = (int(_pses[kx].hifreq) * fmult) / 1000000.0
            fhigh = round(fhigh, 4)
            if (flow > 0) and (flow >= fhigh):   # reverse em
                val = flow
                flow = fhigh
                fhigh = val
            rx = RadioSettingValueFloat(0, 550.0, flow, 0.010, 3)
            rset = RadioSetting("pgmscanedge/%d.lofreq" % kx,
                                "-- Scan %d Low Limit" % kx, rx)
            rset.set_apply_callback(myset_frqflgs, _pses, kx, "loflags",
                                    "lofreq")
            edges.append(rset)

            rx = RadioSettingValueFloat(0, 550.0, fhigh, 0.010, 3)
            rset = RadioSetting("pgmscanedge/%d.hifreq" % kx,
                                "-- Scan %d High Limit" % kx, rx)
            rset.set_apply_callback(myset_frqflgs, _pses, kx, "hiflags",
                                    "hifreq")
            edges.append(rset)

            # Tstep and Mode depend on the bands
            ndxt = 0
            ndxm = 0
            bxnd = 0
            tsopt = ["-", "5k", "6.25k", "10k", "12.5k", "15k",
                     "20k", "25k", "30k", "50k"]
            mdopt = ["-", "FM", "FM-N"]
            if fhigh > 0:
                if fhigh < 135.0:  # Air band
                    bxnd = 1
                    tsopt = ["-", "8.33k", "25k", "Auto"]
                    ndxt = _pses[kx].tstp
                    if ndxt == 0xe:     # Auto
                        ndxt = 3
                    elif ndxt == 8:     # 25k
                        ndxt = 2
                    elif ndxt == 2:     # 8.33k
                        ndxt = 1
                    else:
                        ndxt = 0
                    mdopt = ["-"]
                elif (flow >= 137.0) and (fhigh <= 174.0):   # VHF
                    ndxt = _pses[kx].tstp - 1
                    ndxm = _pses[kx].mode + 1
                    bxnd = 2
                elif (flow >= 375.0) and (fhigh <= 550.0):  # UHF
                    ndxt = _pses[kx].tstp - 1
                    ndxm = _pses[kx].mode + 1
                    bxnd = 3
                else:   # Mixed, ndx's = 0 default
                    tsopt = ["-"]
                    mdopt = ["-"]
                    bxnd = 4
                if (ndxt > 9) or (ndxt < 0):
                    ndxt = 0   # trap ff
                if ndxm > 2:
                    ndxm = 0
            # end if fhigh > 0
            rx = RadioSettingValueList(tsopt, tsopt[ndxt])
            rset = RadioSetting("pgmscanedge/%d.tstp" % kx,
                                "-- Scan %d Freq Step" % kx, rx)
            rset.set_apply_callback(myset_tsopt, _pses, kx, "tstp", bxnd)
            edges.append(rset)

            rx = RadioSettingValueList(mdopt, mdopt[ndxm])
            rset = RadioSetting("pgmscanedge/%d.mode" % kx,
                                "-- Scan %d Mode" % kx, rx)
            rset.set_apply_callback(myset_mdopt, _pses, kx, "mode", bxnd)
            edges.append(rset)
        # End for kx

        # --- Program Scan Links
        _psln = self._memobj.pslnam
        _pslg = self._memobj.pslgrps
        for kx in range(0, 10):
            stx = ""
            for i in range(0, 6):
                stx += chr(int(_psln[kx].nam[i]))
            stx = stx.rstrip()
            rx = RadioSettingValueString(0, 6, stx)
            rset = RadioSetting("pslnam/%d.nam" % kx,
                                "Program Scan Link %d Name" % kx, rx)
            rset.set_apply_callback(myset_psnam, _psln, kx, "nam", 6)
            pslnk.append(rset)

            for px in range(0, 25):
                # Generate string numeric representation of 4-byte bitmask
                stx = ""
                for nx in range(0, 25):
                    if nx < 8:
                        if (_pslg[kx].msk[0] & (1 << nx)):
                            stx += "%0d, " % nx
                    elif (nx >= 8) and (nx < 16):
                        if (_pslg[kx].msk[1] & (1 << (nx - 8))):
                            stx += "%0d, " % nx
                    elif (nx >= 16) and (nx < 24):
                        if (_pslg[kx].msk[2] & (1 << (nx - 16))):
                            stx += "%0d, " % nx
                    elif (nx >= 24):
                        if (_pslg[kx].msk[3] & (1 << (nx - 24))):
                            stx += "%0d, " % nx
            rx = RadioSettingValueString(0, 80, stx)
            rset = RadioSetting("pslgrps/%d.msk" % kx,
                                "--- Scan Link %d Scans" % kx, rx)
            rset.set_apply_callback(myset_bitmask, _pslg,
                                    kx, "msk", 25)
            pslnk.append(rset)
            # end for px
        # End for kx
        return group       # END get_settings()

    def set_settings(self, settings):
        _settings = self._memobj.settings
        _mem = self._memobj
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif element.get_name() == "drv_clone_speed":
                        val = element.value.get_value()
                        self.__class__._can_hispeed = val
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise



# Copyright 2016 Pavel Milanes CO7WT, <co7wt@frcuba.co.cu> <pavelmc@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import struct
import logging

LOG = logging.getLogger(__name__)

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettings

MEM_FORMAT = """
#seekto 0x0010;
struct {
  lbcd rxfreq[4];
  lbcd txfreq[4];
  lbcd rx_tone[2];
  lbcd tx_tone[2];
  u8 unknown41:1,
     unknown42:1,
     power:1,           // high power set (1=off)
     shift:1,           // Shift (1=off)
     busy:1,            // Busy lock (1=off)
     unknown46:1,
     unknown47:1,
     unknown48:1;
  u8 rxen;              // xff if off, x00 if enabled (if chan sel = 00)
  u8 txen;              // xff if off, x00 if enabled
  u8 unknown7;
} memory[32];

#seekto 0x0338;
u8 scan[4];             //  4 bytes / bit LSBF for the channel

#seekto 0x033C;
u8 active[4];           //  4 bytes / bit LSBF for the active cha
                        // active = 0

#seekto 0x0340;
struct {
  u8 kMoni;             // monitor key function
  u8 kScan;             // scan key function
  u8 kDial;             // dial key function
  u8 kTa;               // ta key function
  u8 kLo;               // low key function
  u8 unknown40[7];
  // 0x034c
  u8 tot;               // TOT val * 30 steps (x00-0xa)
  u8 tot_alert;         // TOT pre-alert val * 10 steps, (x00-x19)
  u8 tot_rekey;         // TOT rekey val, 0-60, (x00-x3c)
  u8 tot_reset;         // TOT reset val, 0-15, (x00-x0f)
  // 0x0350
  u8 sql;               // SQL level val, 0-9 (default 6)
  u8 unknown50[12];
  u8 unknown30:1,
     unknown31:1,
     dealer:1,          // dealer & test mode (1=on)
     add:1,             // add/del from the scan (1=on)
     unknown34:1,
     batt_save:1,       // Battery save (1=on)
     unknown36:1,
     beep:1;            // beep on tone (1=on)
  u8 unknown51[2];
} settings;

#seekto 0x03f0;
struct {
  u8 batt_level;        // inverted (ff-val)
  u8 sq_tight;          // sq tight (ff-val)
  u8 sq_open;           // sq open (ff-val)
  u8 high_power;        // High power
  u8 qt_dev;            // QT deviation
  u8 dqt_dev;           // DQT deviation
  u8 low_power;         // low power
} tune;

"""

MEM_SIZE = 0x400
BLOCK_SIZE = 8
MEM_BLOCKS = list(range(0, (MEM_SIZE // BLOCK_SIZE)))
ACK_CMD = "\x06"
TIMEOUT = 0.05  # from 0.03 up it' s safe, we set in 0.05 for a margin

POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=1),
                chirp_common.PowerLevel("High", watts=5)]
SKIP_VALUES = ["", "S"]
TONES = chirp_common.TONES
# TONES.remove(254.1)
DTCS_CODES = chirp_common.DTCS_CODES

# some vars for the UI
off = ["off"]
TOT = off + ["%s" % x for x in range(30, 330, 30)]
TOT_A = off + ["%s" % x for x in range(10, 260, 10)]
TOT_RK = off + ["%s" % x for x in range(1, 61)]
TOT_RS = off + ["%s" % x for x in range(1, 16)]
SQL = off + ["%s" % x for x in range(1, 10)]

# keys
MONI = off + ["Monitor momentary", "Monitor lock", "SQ off momentary"]
SCAN = off + ["Carrier operated (COS)", "Time operated (TOS)"]
YESNO = ["Enabled", "Disabled"]
TA = off + ["Turn around", "Reverse"]


def rawrecv(radio, amount):
    """Raw read from the radio device"""
    data = ""
    try:
        data = radio.pipe.read(amount)
        # print("<= %02i: %s" % (len(data), util.hexprint(data)))
    except:
        raise errors.RadioError("Error reading data from radio")

    return data


def rawsend(radio, data):
    """Raw send to the radio device"""
    try:
        radio.pipe.write(data)
        # print("=> %02i: %s" % (len(data), util.hexprint(data)))
    except:
        raise errors.RadioError("Error sending data from radio")


def send(radio, frame):
    """Generic send data to the radio"""
    rawsend(radio, frame)


def make_frame(cmd, addr, data=""):
    """Pack the info in the format it likes"""
    ts = struct.pack(">BHB", ord(cmd), addr, 8)
    if data == "":
        return ts
    else:
        if len(data) == 8:
            return ts + data
        else:
            raise errors.InvalidValueError("Data of unexpected length to send")


def handshake(radio, msg="", full=False):
    """Make a full handshake, if not full just hals"""
    # send ACK if commandes
    if full is True:
        rawsend(radio, ACK_CMD)
    # receive ACK
    ack = rawrecv(radio, 1)
    # check ACK
    if ack != ACK_CMD:
        # close_radio(radio)
        mesg = "Handshake failed: " + msg
        raise Exception(mesg)


def recv(radio):
    """Receive data from the radio, 12 bytes, 4 in the header, 8 as data"""
    rxdata = rawrecv(radio, 12)

    if len(rxdata) != 12:
        raise errors.RadioError(
            "Received a length of data that is not possible")

    cmd, addr, length = struct.unpack(">BHB", rxdata[0:4])
    data = ""
    if length == 8:
        data = rxdata[4:]

    return data


def open_radio(radio):
    """Open the radio into program mode and check if it's the correct model"""
    # Set serial discipline
    try:
        radio.pipe.parity = "N"
        radio.pipe.timeout = TIMEOUT
        radio.pipe.flush()
    except:
        msg = "Serial error: Can't set serial line discipline"
        raise errors.RadioError(msg)

    # we will try to open the radio 5 times, this is an improved mechanism
    magic = "PROGRAM"
    exito = False
    for i in range(0, 5):
        for i in range(0, len(magic)):
            ack = rawrecv(radio, 1)
            time.sleep(0.05)
            send(radio, magic[i])

        try:
            handshake(radio, "Radio not entering Program mode")
            exito = True
            break
        except:
            LOG.debug("Attempt #%s, failed, trying again" % i)
            pass

    # check if we had EXITO
    if exito is False:
        msg = "The radio did not accept program mode after five tries.\n"
        msg += "Check you interface cable and power cycle your radio."
        raise errors.RadioError(msg)

    rawsend(radio, "\x02")
    ident = rawrecv(radio, 8)
    handshake(radio, "Comm error after ident", True)

    if not (radio.TYPE in ident):
        LOG.debug("Incorrect model ID, got %s" % util.hexprint(ident))
        msg = "Incorrect model ID, got %s, it not contains %s" % \
            (ident[0:5], radio.TYPE)
        raise errors.RadioError(msg)


def do_download(radio):
    """This is your download function"""
    open_radio(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE / BLOCK_SIZE
    status.msg = "Cloning from radio..."
    radio.status_fn(status)

    data = ""
    for addr in MEM_BLOCKS:
        send(radio, make_frame("R", addr * BLOCK_SIZE))
        data += recv(radio)
        handshake(radio, "Rx error in block %03i" % addr, True)
        # DEBUG
        # print("Block: %04x, Pos: %06x" % (addr, addr * BLOCK_SIZE))

        # UI Update
        status.cur = addr
        status.msg = "Cloning from radio..."
        radio.status_fn(status)

    return memmap.MemoryMap(data)


def do_upload(radio):
    """Upload info to radio"""
    open_radio(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE / BLOCK_SIZE
    status.msg = "Cloning to radio..."
    radio.status_fn(status)
    count = 0

    for addr in MEM_BLOCKS:
        # UI Update
        status.cur = addr
        status.msg = "Cloning to radio..."
        radio.status_fn(status)

        block = addr * BLOCK_SIZE
        # Beyond 0x03d0 the data is not writable
        if block > 0x3d0:
            continue

        data = radio.get_mmap()[block:block + BLOCK_SIZE]
        send(radio, make_frame("W", block, data))

        time.sleep(0.02)
        handshake(radio, "Rx error in block %03i" % addr)


def get_radio_id(data):
    """Extract the radio identification from the firmware"""
    # Reverse the radio id string. MemoryMap does not support the step/stride
    # slice argument, so it is first sliced to a str then reversed.
    return data[0x03d0:0x03d8][::-1]


def model_match(cls, data):
    """Match the opened/downloaded image to the correct version"""
    rid = get_radio_id(data)

    # DEBUG
    # print("Full ident string is %s" % util.hexprint(rid))

    if (rid in cls.VARIANTS):
        # correct model
        return True
    else:
        return False


class Kenwood_P60_Radio(chirp_common.CloneModeRadio, chirp_common.ExperimentalRadio):
    """Kenwood Mobile Family 60 Radios"""
    VENDOR = "Kenwood"
    _range = [350000000, 512000000]  # don't mind, it will be overited
    _upper = 32
    VARIANT = ""
    MODEL = ""
    _kind = ""

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('This driver is experimental; not all features have been '
             'implemented, but it has those features most used by hams.\n'
             '\n'
             'This radios are able to work slightly outside the OEM '
             'frequency limits. After testing, the limit in Chirp has '
             'been set 4% outside the OEM limit. This allows you to use '
             'some models on the ham bands.\n'
             '\n'
             'Nevertheless, each radio has its own hardware limits and '
             'your mileage may vary.\n'
             )
        rp.pre_download = _(
            "Follow this instructions to read your radio:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the download of your radio data\n")
        rp.pre_upload = _(
            "Follow this instructions to write your radio:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the upload of your radio data\n")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.has_name = False
        rf.has_offset = True
        rf.has_mode = False
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_modes = ["FM"]
        rf.valid_duplexes = ["", "-", "+", "off"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_skips = SKIP_VALUES
        rf.valid_dtcs_codes = DTCS_CODES
        rf.valid_bands = [self._range]
        rf.memory_bounds = (1, self._upper)
        rf.valid_tuning_steps = [2.5, 5.0, 6.25, 10.0, 12.5, 25.0]
        return rf

    def sync_in(self):
        """Download from radio"""
        self._mmap = do_download(self)
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        # Get the data ready for upload
        try:
            self._prep_data()
        except:
            raise errors.RadioError("Error processing the radio data")

        # do the upload
        try:
            do_upload(self)
        except:
            raise errors.RadioError("Error uploading data to radio")

    def set_variant(self):
        """Select and set the correct variables for the class according
        to the correct variant of the radio"""
        rid = get_radio_id(self._mmap)

        # identify the radio variant and set the environment to its values
        try:
            self._upper, low, high, self._kind = self.VARIANTS[rid]

            # Frequency ranges: some model/variants are able to work the near
            # ham bands, even if they are outside the OEM ranges.
            # By experimentation we found that a +/- 4% at the edges is in most
            # cases safe and will cover the near ham band in full
            self._range = [low * 1000000 * 0.96, high * 1000000 * 1.04]

            # put the VARIANT in the class, clean the model / CHs / Type
            # in the same layout as the KPG program
            self._VARIANT = self.MODEL + " [" + str(self._upper) + "CH]: "
            # In the OEM string we show the real OEM ranges
            self._VARIANT += self._kind + ", %d - %d MHz" % (low, high)

            # DEBUG
            # print self._VARIANT

        except KeyError:
            LOG.debug("Wrong Kenwood radio, ID or unknown variant")
            LOG.debug(util.hexprint(rid))
            raise errors.RadioError(
                "Wrong Kenwood radio, ID or unknown variant, see LOG output.")

    def _prep_data(self):
        """Prepare the areas in the memmap to do a consistent write
        it has to make an update on the x280 flag data"""
        achs = 0

        for i in range(0, self._upper):
            if self.get_active(i) is True:
                achs += 1

        # The x0280 area has the settings for the DTMF/2-Tone per channel,
        # as we don't support this feature yet,
        # we disabled by cleaning the data
        # fldata = "\x00\xf0\xff\xff\xff" * achs + \
            # "\xff" * (5 * (self._upper - achs))

        fldata = "\xFF" * 5 * self._upper
        self._fill(0x0280, fldata)

    def _fill(self, offset, data):
        """Fill an specified area of the memmap with the passed data"""
        for addr in range(0, len(data)):
            self._mmap[offset + addr] = data[addr]

    def process_mmap(self):
        """Process the mem map into the mem object"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)
        # to set the vars on the class to the correct ones
        self.set_variant()

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def get_active(self, chan):
        """Get the channel active status from the 4 bytes array on the eeprom"""
        byte = int(chan/8)
        bit = chan % 8
        res = self._memobj.active[byte] & (pow(2, bit))
        res = not bool(res)

        return res

    def set_active(self, chan, value=True):
        """Set the channel active status from UI to the mem_map"""
        byte = int(chan/8)
        bit = chan % 8

        # DEBUG
        # print("SET Chan %s, Byte %s, Bit % s" % (chan, byte, bit))

        # get the actual value to see if I need to change anything
        actual = self.get_active(chan)
        if actual != bool(value):
            # DEBUG
            # print "VALUE %s flipping" % int(not value)

            # I have to flip the value
            rbyte = self._memobj.active[byte]
            rbyte = rbyte ^ pow(2, bit)
            self._memobj.active[byte] = rbyte

    def decode_tone(self, val):
        """Parse the tone data to decode from mem, it returns:
        Mode (''|DTCS|Tone), Value (None|###), Polarity (None,N,R)"""
        if val.get_raw(asbytes=False) == "\xFF\xFF":
            return '', None, None

        val = int(val)
        if val >= 12000:
            a = val - 12000
            return 'DTCS', a, 'R'
        elif val >= 8000:
            a = val - 8000
            return 'DTCS', a, 'N'
        else:
            a = val / 10.0
            return 'Tone', a, None

    def encode_tone(self, memval, mode, value, pol):
        """Parse the tone data to encode from UI to mem"""
        if mode == '':
            memval[0].set_raw(0xFF)
            memval[1].set_raw(0xFF)
        elif mode == 'Tone':
            memval.set_value(int(value * 10))
        elif mode == 'DTCS':
            flag = 0x80 if pol == 'N' else 0xC0
            memval.set_value(value)
            memval[1].set_bits(flag)
        else:
            raise Exception("Internal error: invalid mode `%s'" % mode)

    def get_scan(self, chan):
        """Get the channel scan status from the 4 bytes array on the eeprom
        then from the bits on the byte, return '' or 'S' as needed"""
        result = "S"
        byte = int(chan/8)
        bit = chan % 8
        res = self._memobj.scan[byte] & (pow(2, bit))
        if res > 0:
            result = ""

        return result

    def set_scan(self, chan, value):
        """Set the channel scan status from UI to the mem_map"""
        byte = int(chan/8)
        bit = chan % 8

        # get the actual value to see if I need to change anything
        actual = self.get_scan(chan)
        if actual != value:
            # I have to flip the value
            rbyte = self._memobj.scan[byte]
            rbyte = rbyte ^ pow(2, bit)
            self._memobj.scan[byte] = rbyte

    def get_memory(self, number):
        """Get the mem representation from the radio image"""
        _mem = self._memobj.memory[number - 1]

        # Create a high-level memory object to return to the UI
        mem = chirp_common.Memory()

        # Memory number
        mem.number = number

        if _mem.get_raw(asbytes=False)[0] == "\xFF":
            mem.empty = True
            # but is not enough, you have to clear the memory in the mmap
            # to get it ready for the sync_out process, just in case
            _mem.set_raw("\xFF" * 16)
            # set the channel to inactive state
            self.set_active(number - 1, False)
            return mem

        # Freq and offset
        mem.freq = int(_mem.rxfreq) * 10
        # tx freq can be blank
        if _mem.get_raw(asbytes=False)[4] == "\xFF" or int(_mem.txen) == 255:
            # TX freq not set
            mem.offset = 0
            mem.duplex = "off"
        else:
            # TX feq set
            offset = (int(_mem.txfreq) * 10) - mem.freq
            if offset < 0:
                mem.offset = abs(offset)
                mem.duplex = "-"
            elif offset > 0:
                mem.offset = offset
                mem.duplex = "+"
            else:
                mem.offset = 0

        # power
        mem.power = POWER_LEVELS[int(_mem.power)]

        # skip
        mem.skip = self.get_scan(number - 1)

        # tone data
        rxtone = txtone = None
        txtone = self.decode_tone(_mem.tx_tone)
        rxtone = self.decode_tone(_mem.rx_tone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        # Extra
        # bank and number in the channel
        mem.extra = RadioSettingGroup("extra", "Extra")

        bl = RadioSetting("busy", "Busy Channel lock",
                          RadioSettingValueBoolean(
                              not bool(_mem.busy)))
        mem.extra.append(bl)

        sf = RadioSetting("shift", "Beat Shift",
                          RadioSettingValueBoolean(
                              not bool(_mem.shift)))
        mem.extra.append(sf)

        return mem

    def set_memory(self, mem):
        """Set the memory data in the eeprom img from the UI
        not ready yet, so it will return as is"""

        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[mem.number - 1]

        # Empty memory
        if mem.empty:
            _mem.set_raw("\xFF" * 16)
            self.set_active(mem.number - 1, False)
            return

        # freq rx
        _mem.rxfreq = mem.freq / 10

        # rx enabled if valid channel,
        # set tx to on, we decide if off after duplex = off
        _mem.rxen = 0
        _mem.txen = 0

        # freq tx
        if mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        elif mem.duplex == "off":
            # set tx freq on the memap to xff
            for i in range(0, 4):
                _mem.txfreq[i].set_raw("\xFF")
            # erase the txen flag
            _mem.txen = 255
        else:
            _mem.txfreq = mem.freq / 10

        # tone data
        ((txmode, txtone, txpol), (rxmode, rxtone, rxpol)) = \
            chirp_common.split_tone_encode(mem)
        self.encode_tone(_mem.tx_tone, txmode, txtone, txpol)
        self.encode_tone(_mem.rx_tone, rxmode, rxtone, rxpol)

        # power, default power is high, as the low is configurable via a key
        if mem.power is None:
            mem.power = POWER_LEVELS[1]

        _mem.power = POWER_LEVELS.index(mem.power)

        # skip
        self.set_scan(mem.number - 1, mem.skip)

        # set as active
        self.set_active(mem.number - 1, True)

        # extra settings
        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

        return mem

    @classmethod
    def match_model(cls, filedata, filename):
        match_size = False
        match_model = False

        # testing the file data size
        if len(filedata) == MEM_SIZE:
            match_size = True

        # testing the firmware model fingerprint
        match_model = model_match(cls, filedata)

        if match_size and match_model:
            return True
        else:
            return False

    def get_settings(self):
        """Translate the bit in the mem_struct into settings in the UI"""
        sett = self._memobj.settings

        # basic features of the radio
        basic = RadioSettingGroup("basic", "Basic Settings")
        # keys
        fkeys = RadioSettingGroup("keys", "Function keys")

        top = RadioSettings(basic, fkeys)

        # Basic
        val = RadioSettingValueString(0, 35, self._VARIANT)
        val.set_mutable(False)
        mod = RadioSetting("not.mod", "Radio version", val)
        basic.append(mod)

        beep = RadioSetting("settings.beep", "Beep tone",
                            RadioSettingValueBoolean(
                                 bool(sett.beep)))
        basic.append(beep)

        bsave = RadioSetting("settings.batt_save", "Battery save",
                             RadioSettingValueBoolean(
                                 bool(sett.batt_save)))
        basic.append(bsave)

        deal = RadioSetting("settings.dealer", "Dealer & Test",
                            RadioSettingValueBoolean(
                                 bool(sett.dealer)))
        basic.append(deal)

        add = RadioSetting("settings.add", "Del / Add feature",
                           RadioSettingValueBoolean(
                                 bool(sett.add)))
        basic.append(add)

        # In some cases the values that follows can be 0xFF (HARD RESET)
        # so we need to take and validate that
        if int(sett.tot) == 0xff:
            # 120 sec
            sett.tot = 4
        if int(sett.tot_alert) == 0xff:
            # 10 secs
            sett.tot_alert = 1
        if int(sett.tot_rekey) == 0xff:
            # off
            sett.tot_rekey = 0
        if int(sett.tot_reset) == 0xff:
            # off
            sett.tot_reset = 0
        if int(sett.sql) == 0xff:
            # a comfortable level ~6
            sett.sql = 6

        tot = RadioSetting("settings.tot", "Time Out Timer (TOT)",
                           RadioSettingValueList(TOT, TOT[int(sett.tot)]))
        basic.append(tot)

        tota = RadioSetting("settings.tot_alert", "TOT pre-plert",
                            RadioSettingValueList(TOT_A, TOT_A[int(sett.tot_alert)]))
        basic.append(tota)

        totrk = RadioSetting("settings.tot_rekey", "TOT rekey time",
                             RadioSettingValueList(TOT_RK, TOT_RK[int(sett.tot_rekey)]))
        basic.append(totrk)

        totrs = RadioSetting("settings.tot_reset", "TOT reset time",
                             RadioSettingValueList(TOT_RS, TOT_RS[int(sett.tot_reset)]))
        basic.append(totrs)

        sql = RadioSetting("settings.sql", "Squelch level",
                           RadioSettingValueList(SQL, SQL[int(sett.sql)]))
        basic.append(sql)

        # front keys
        m = int(sett.kMoni)
        if m > 3:
            m = 1
        mon = RadioSetting("settings.kMoni", "Monitor",
                           RadioSettingValueList(MONI, MONI[m]))
        fkeys.append(mon)

        s = int(sett.kScan)
        if s > 3:
            s = 1
        scn = RadioSetting("settings.kScan", "Scan",
                           RadioSettingValueList(SCAN, SCAN[s]))
        fkeys.append(scn)

        d = int(sett.kDial)
        if d > 1:
            d = 0
        dial = RadioSetting("settings.kDial", "Dial",
                            RadioSettingValueList(YESNO, YESNO[d]))
        fkeys.append(dial)

        t = int(sett.kTa)
        if t > 2:
            t = 2
        ta = RadioSetting("settings.kTa", "Ta",
                          RadioSettingValueList(TA, TA[t]))
        fkeys.append(ta)

        l = int(sett.kLo)
        if l > 1:
            l = 0
        low = RadioSetting("settings.kLo", "Low",
                           RadioSettingValueList(YESNO, YESNO[l]))
        fkeys.append(low)

        return top

    def set_settings(self, settings):
        """Translate the settings in the UI into bit in the mem_struct
        I don't understand well the method used in many drivers
        so, I used mine, ugly but works ok"""

        mobj = self._memobj

        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue

            # Let's roll the ball
            if "." in element.get_name():
                inter, setting = element.get_name().split(".")
                # you must ignore the settings with "not"
                # this are READ ONLY attributes
                if inter == "not":
                    continue

                obj = getattr(mobj, inter)
                value = element.value

                # integers case + special case
                if setting in ["tot", "tot_alert", "tot_rekey",
                               "tot_reset", "sql", "kMoni", "kScan",
                               "kDial", "kTa", "kLo"]:
                    # catching the "off" values as zero
                    try:
                        value = int(value)
                    except:
                        value = 0

                # Bool types + inverted
                if setting in ["beep", "batt_save", "dealer", "add"]:
                    value = bool(value)

            # Apply al configs done
            # DEBUG
            # print("%s: %s" % (setting, value))
            setattr(obj, setting, value)


# This are the oldest family 60 models just portables support here
# Info striped from a hexdump inside the preogram and hack over a
# tk-270

@directory.register
class TK260_Radio(Kenwood_P60_Radio):
    """Kenwood TK-260 Radios"""
    MODEL = "TK-260"
    TYPE = "P0260"
    VARIANTS = {
        b"P0260\x20\x00\x00": (4, 136, 150, "F2"),
        b"P0260\x21\x00\x00": (4, 150, 174, "F1"),
        }


@directory.register
class TK270_Radio(Kenwood_P60_Radio):
    """Kenwood TK-270 Radios"""
    MODEL = "TK-270"
    TYPE = "P0270"
    VARIANTS = {
        b"P0270\x10\x00\x00": (32, 136, 150, "F2"),
        b"P0270\x11\x00\x00": (32, 150, 174, "F1"),
        }


@directory.register
class TK272_Radio(Kenwood_P60_Radio):
    """Kenwood TK-272 Radios"""
    MODEL = "TK-272"
    TYPE = "P0272"
    VARIANTS = {
        b"P0272\x10\x00\x00": (10, 136, 150, "F2"),
        b"P0272\x11\x00\x00": (10, 150, 174, "F1"),
        }


@directory.register
class TK278_Radio(Kenwood_P60_Radio):
    """Kenwood TK-278 Radios"""
    MODEL = "TK-278"
    TYPE = "P0278"
    VARIANTS = {
        b"P0278\x00\x00\x00": (32, 136, 150, "F2"),
        b"P0278\x01\x00\x00": (32, 150, 174, "F1"),
        }


@directory.register
class TK360_Radio(Kenwood_P60_Radio):
    """Kenwood TK-360 Radios"""
    MODEL = "TK-360"
    TYPE = "P0360"
    VARIANTS = {
        b"P0360\x24\x00\x00": (4, 450, 470, "F1"),
        b"P0360\x25\x00\x00": (4, 470, 490, "F2"),
        b"P0360\x26\x00\x00": (4, 490, 512, "F3"),
        b"P0360\x23\x00\x00": (4, 406, 430, "F4"),
        }


@directory.register
class TK370_Radio(Kenwood_P60_Radio):
    """Kenwood TK-370 Radios"""
    MODEL = "TK-370"
    TYPE = "P0370"
    VARIANTS = {
        b"P0370\x14\x00\x00": (32, 450, 470, "F1"),
        b"P0370\x15\x00\x00": (32, 470, 490, "F2"),
        b"P0370\x16\x00\x00": (32, 490, 512, "F3"),
        b"P0370\x13\x00\x00": (32, 406, 430, "F4"),
        }


@directory.register
class TK372_Radio(Kenwood_P60_Radio):
    """Kenwood TK-372 Radios"""
    MODEL = "TK-372"
    TYPE = "P0372"
    VARIANTS = {
        b"P0372\x14\x00\x00": (10, 450, 470, "F1"),
        b"P0372\x15\x00\x00": (10, 470, 490, "F2"),
        }


@directory.register
class TK378_Radio(Kenwood_P60_Radio):
    """Kenwood TK-378 Radios"""
    MODEL = "TK-378"
    TYPE = "P0378"
    VARIANTS = {
        b"P0378\x04\x00\x00": (32, 370, 470, "SP1"),
        b"P0378\x02\x00\x00": (32, 350, 427, "SP2"),
        }



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp.drivers import icf
from chirp import chirp_common, util, directory, bitwise

MEM_FORMAT = """
struct {
  u32  freq;
  u32  offset;
  char urcall[8];
  char r1call[8];
  char r2call[8];
  u8   unknown1;
  u8   unknown2:1,
       duplex:2,
       tmode:3,
       unknown3:2;
  u16  ctone:6,
       rtone:6,
       tune_step:4;
  u16  dtcs:7,
       mode:3,
       unknown4:6;
  u8   unknown5:1,
       digital_code:7;
  u8   unknown6:2,
       dtcs_polarity:2,
       unknown7:4;
  char name[8];
} memory[522];

#seekto 0x61E0;
u8 used_flags[66];

#seekto 0x6222;
u8 skip_flags[65];
u8 pskip_flags[65];

#seekto 0x62A4;
struct {
  u8 bank;
  u8 index;
} bank_info[500];

#seekto 0x66C0;
struct {
  char name[8];
} bank_names[26];

#seekto 0x6970;
struct {
  char call[8];
  u8 unknown[4];
} mycall[6];

#seekto 0x69B8;
struct {
  char call[8];
} urcall[60];

struct {
  char call[8];
} rptcall[60];

"""

TMODES = ["", "Tone", "??0", "TSQL", "??1", "??2", "DTCS"]
DUPLEX = ["", "-", "+", "+"]  # Not sure about index 3
MODES = ["FM", "NFM", "AM", "??", "DV"]
DTCSP = ["NN", "NR", "RN", "RR"]

MEM_LOC_SIZE = 48


class IC2820Bank(icf.IcomNamedBank):
    """An IC2820 bank"""
    def get_name(self):
        _banks = self._model._radio._memobj.bank_names
        return str(_banks[self.index].name).rstrip()

    def set_name(self, name):
        _banks = self._model._radio._memobj.bank_names
        _banks[self.index].name = str(name).ljust(8)[:8]


def _get_special():
    special = {"C0": 500 + 20,
               "C1": 500 + 21}

    for i in range(0, 10):
        ida = "%iA" % i
        idb = "%iB" % i
        special[ida] = 500 + i * 2
        special[idb] = 500 + i * 2 + 1

    return special


def _resolve_memory_number(number):
    if isinstance(number, str):
        return _get_special()[number]
    else:
        return number


def _wipe_memory(mem, char):
    mem.set_raw(char * (mem.size() // 8))


@directory.register
class IC2820Radio(icf.IcomCloneModeRadio, chirp_common.IcomDstarSupport):
    """Icom IC-2820"""
    VENDOR = "Icom"
    MODEL = "IC-2820H"

    _model = "\x29\x70\x00\x01"
    _memsize = 44224
    _endframe = "Icom Inc\x2e68"

    _ranges = [(0x0000, 0x6960, 32),
               (0x6960, 0x6980, 16),
               (0x6980, 0x7160, 32),
               (0x7160, 0x7180, 16),
               (0x7180, 0xACC0, 32),
               ]

    _num_banks = 26
    _bank_class = IC2820Bank
    _can_hispeed = True

    MYCALL_LIMIT = (1, 7)
    URCALL_LIMIT = (1, 61)
    RPTCALL_LIMIT = (1, 61)

    _memories = {}

    def _get_bank(self, loc):
        _bank = self._memobj.bank_info[loc]
        if _bank.bank == 0xFF:
            return None
        else:
            return _bank.bank

    def _set_bank(self, loc, bank):
        _bank = self._memobj.bank_info[loc]
        if bank is None:
            _bank.bank = 0xFF
        else:
            _bank.bank = bank

    def _get_bank_index(self, loc):
        _bank = self._memobj.bank_info[loc]
        return _bank.index

    def _set_bank_index(self, loc, index):
        _bank = self._memobj.bank_info[loc]
        _bank.index = index

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank_index = True
        rf.has_bank_names = True
        rf.requires_call_lists = False
        rf.memory_bounds = (0, 499)
        rf.valid_modes = [x for x in MODES if '?' not in x]
        rf.valid_tmodes = list(TMODES)
        rf.valid_duplexes = list(set(DUPLEX))
        rf.valid_tuning_steps = list(chirp_common.TUNING_STEPS)
        rf.valid_bands = [(118000000, 999990000)]
        rf.valid_skips = ["", "S", "P"]
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.valid_name_length = 8
        rf.valid_special_chans = sorted(_get_special().keys())

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_memory(self, number):
        number = _resolve_memory_number(number)

        bitpos = (1 << (number % 8))
        bytepos = number / 8

        _mem = self._memobj.memory[number]
        _used = self._memobj.used_flags[bytepos]

        is_used = ((_used & bitpos) == 0)

        if is_used and MODES[_mem.mode] == "DV":
            mem = chirp_common.DVMemory()
            mem.dv_urcall = str(_mem.urcall).rstrip()
            mem.dv_rpt1call = str(_mem.r1call).rstrip()
            mem.dv_rpt2call = str(_mem.r2call).rstrip()
        else:
            mem = chirp_common.Memory()

        mem.number = number
        if number < 500:
            _skip = self._memobj.skip_flags[bytepos]
            _pskip = self._memobj.pskip_flags[bytepos]
            if _skip & bitpos:
                mem.skip = "S"
            elif _pskip & bitpos:
                mem.skip = "P"
        else:
            mem.extd_number = util.get_dict_rev(_get_special(), number)
            mem.immutable = ["number", "skip", "extd_number"]

        if not is_used:
            mem.empty = True
            return mem

        mem.freq = int(_mem.freq)
        mem.offset = int(_mem.offset)
        mem.rtone = chirp_common.TONES[_mem.rtone]
        mem.ctone = chirp_common.TONES[_mem.ctone]
        mem.tmode = TMODES[_mem.tmode]
        mem.duplex = DUPLEX[_mem.duplex]
        mem.mode = MODES[_mem.mode]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.dtcs_polarity = DTCSP[_mem.dtcs_polarity]
        if _mem.tune_step > 8:
            mem.tuning_step = 5.0  # Sometimes TS is garbage?
        else:
            mem.tuning_step = chirp_common.TUNING_STEPS[_mem.tune_step]
        mem.name = str(_mem.name).rstrip()

        return mem

    def set_memory(self, mem):
        bitpos = (1 << (mem.number % 8))
        bytepos = mem.number / 8

        _mem = self._memobj.memory[mem.number]
        _used = self._memobj.used_flags[bytepos]

        was_empty = _used & bitpos

        if mem.number < 500:
            skip = self._memobj.skip_flags[bytepos]
            pskip = self._memobj.pskip_flags[bytepos]
            if mem.skip == "S":
                skip |= bitpos
            else:
                skip &= ~bitpos
            if mem.skip == "P":
                pskip |= bitpos
            else:
                pskip &= ~bitpos

        if mem.empty:
            _used |= bitpos
            _wipe_memory(_mem, "\xFF")
            if mem.number < 500:
                self._set_bank(mem.number, None)
            return

        _used &= ~bitpos
        if was_empty:
            _wipe_memory(_mem, "\x00")

        _mem.freq = mem.freq
        _mem.offset = mem.offset
        _mem.rtone = chirp_common.TONES.index(mem.rtone)
        _mem.ctone = chirp_common.TONES.index(mem.ctone)
        _mem.tmode = TMODES.index(mem.tmode)
        _mem.duplex = DUPLEX.index(mem.duplex)
        _mem.mode = MODES.index(mem.mode)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.dtcs_polarity = DTCSP.index(mem.dtcs_polarity)
        _mem.tune_step = chirp_common.TUNING_STEPS.index(mem.tuning_step)
        _mem.name = mem.name.ljust(8)

        if isinstance(mem, chirp_common.DVMemory):
            _mem.urcall = mem.dv_urcall.ljust(8)
            _mem.r1call = mem.dv_rpt1call.ljust(8)
            _mem.r2call = mem.dv_rpt2call.ljust(8)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def get_urcall_list(self):
        _calls = self._memobj.urcall
        calls = []

        for i in range(*self.URCALL_LIMIT):
            calls.append(str(_calls[i-1].call))

        return calls

    def get_repeater_call_list(self):
        _calls = self._memobj.rptcall
        calls = []

        for i in range(*self.RPTCALL_LIMIT):
            calls.append(str(_calls[i-1].call))

        return calls

    def get_mycall_list(self):
        _calls = self._memobj.mycall
        calls = []

        for i in range(*self.MYCALL_LIMIT):
            calls.append(str(_calls[i-1].call))

        return calls

    def set_urcall_list(self, calls):
        _calls = self._memobj.urcall

        for i in range(*self.URCALL_LIMIT):
            try:
                call = calls[i-1]
            except IndexError:
                call = " " * 8

            _calls[i-1].call = call.ljust(8)[:8]

    def set_repeater_call_list(self, calls):
        _calls = self._memobj.rptcall

        for i in range(*self.RPTCALL_LIMIT):
            try:
                call = calls[i-1]
            except IndexError:
                call = " " * 8

            _calls[i-1].call = call.ljust(8)[:8]

    def set_mycall_list(self, calls):
        _calls = self._memobj.mycall

        for i in range(*self.MYCALL_LIMIT):
            try:
                call = calls[i-1]
            except IndexError:
                call = " " * 8

            _calls[i-1].call = call.ljust(8)[:8]



# Copyright 2013 Sean Burford <sburford@google.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from chirp import bandplan, bandplan_iaru_r3


SHORTNAME = "australia"

DESC = {
  "name":    "Australian Amateur Band Plan",
  "updated": "April 2010",
  "url":     "http://www.wia.org.au/members/bandplans/data"
             "/documents/Australian%20Band%20Plans%20100404.pdf",
}

BANDS_10M = (
  # bandplan.Band((28000000, 29700000), "10 Meter Band"),
  bandplan.Band((29520000, 29680000), "FM Simplex and Repeaters",
                mode="FM", step_khz=20),
  bandplan.Band((29620000, 29680000), "FM Repeaters", input_offset=-100000),
)

BANDS_6M = (
  # bandplan.Band((50000000, 54000000), "6 Meter Band"),
  bandplan.Band((52525000, 53975000), "FM Simplex and Repeaters",
                mode="FM", step_khz=25),
  bandplan.Band((53550000, 53975000), "FM Repeaters", input_offset=-1000000),
)

BANDS_2M = (
  bandplan.Band((144000000, 148000000), "2 Meter Band",
                tones=(91.5, 123.0, 141.3, 146.2, 85.4)),
  bandplan.Band((144400000, 144600000), "Beacons", step_khz=1),
  bandplan.Band((146025000, 147975000), "FM Simplex and Repeaters",
                mode="FM", step_khz=25),
  bandplan.Band((146625000, 147000000), "FM Repeaters Group A",
                input_offset=-600000),
  bandplan.Band((147025000, 147375000), "FM Repeaters Group B",
                input_offset=600000),
)

BANDS_70CM = (
  bandplan.Band((420000000, 450000000), "70cm Band",
                tones=(91.5, 123.0, 141.3, 146.2, 85.4)),
  bandplan.Band((432400000, 432600000), "Beacons", step_khz=1),
  bandplan.Band((438025000, 439975000), "FM Simplex and Repeaters",
                mode="FM", step_khz=25),
  bandplan.Band((438025000, 438725000), "FM Repeaters Group A",
                input_offset=-5000000),
  bandplan.Band((439275000, 439975000), "FM Repeaters Group B",
                input_offset=-5000000),
)

BANDS_23CM = (
  # bandplan.Band((1240000000, 1300000000), "23cm Band"),
  bandplan.Band((1273025000, 1273975000), "FM Repeaters",
                mode="FM", step_khz=25, input_offset=20000000),
  bandplan.Band((1296400000, 1296600000), "Beacons", step_khz=1),
  bandplan.Band((1297025000, 1300400000), "General FM Simplex Data",
                mode="FM", step_khz=25),
)

BANDS_13CM = (
  bandplan.Band((2300000000, 2450000000), "13cm Band"),
  bandplan.Band((2403400000, 2403600000), "Beacons", step_khz=1),
  bandplan.Band((2425000000, 2428000000), "FM Simplex",
                mode="FM", step_khz=25),
  bandplan.Band((2428025000, 2429000000), "FM Duplex (Voice)",
                mode="FM", step_khz=25, input_offset=20000000),
  bandplan.Band((2429000000, 2429975000), "FM Duplex (Data)",
                mode="FM", step_khz=100, input_offset=20000000),
)

BANDS_9CM = (
  bandplan.Band((3300000000, 3600000000), "9cm Band"),
  bandplan.Band((3320000000, 3340000000), "WB Channel 2: Voice/Data",
                step_khz=100),
  bandplan.Band((3400400000, 3400600000), "Beacons", step_khz=1),
  bandplan.Band((3402000000, 3403000000), "FM Simplex (Voice)",
                mode="FM", step_khz=100),
  bandplan.Band((3403000000, 3405000000), "FM Simplex (Data)",
                mode="FM", step_khz=100),
)

BANDS_6CM = (
  bandplan.Band((5650000000, 5850000000), "6cm Band"),
  bandplan.Band((5760400000, 5760600000), "Beacons", step_khz=1),
  bandplan.Band((5700000000, 5720000000), "WB Channel 2: Data",
                step_khz=100, input_offset=70000000),
  bandplan.Band((5720000000, 5740000000), "WB Channel 3: Voice",
                step_khz=100, input_offset=70000000),
  bandplan.Band((5762000000, 5763000000), "FM Simplex (Voice)",
                mode="FM", step_khz=100),
  bandplan.Band((5763000000, 5765000000), "FM Simplex (Data)",
                mode="FM", step_khz=100),
)

BANDS = bandplan_iaru_r3.BANDS_20M + bandplan_iaru_r3.BANDS_17M + \
        bandplan_iaru_r3.BANDS_15M + bandplan_iaru_r3.BANDS_12M + \
        bandplan_iaru_r3.BANDS_10M + bandplan_iaru_r3.BANDS_6M + \
        BANDS_10M + BANDS_6M + BANDS_2M + BANDS_70CM + BANDS_23CM + \
        BANDS_13CM + BANDS_9CM + BANDS_6CM
